from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_contract.arc import try_isaread_contract_from_path
from ..arctrl_contract.arc_assay import (ARCtrl_ArcAssay__ArcAssay_ToDeleteContract, ARCtrl_ArcAssay__ArcAssay_ToCreateContract_6FCE9E49, ARCtrl_ArcAssay__ArcAssay_ToUpdateContract, ARCtrl_ArcAssay__ArcAssay_tryFromReadContract_Static_7570923F)
from ..arctrl_contract.arc_investigation import (ARCtrl_ArcInvestigation__ArcInvestigation_ToUpdateContract, ARCtrl_ArcInvestigation__ArcInvestigation_tryFromReadContract_Static_7570923F)
from ..arctrl_contract.arc_run import (ARCtrl_ArcRun__ArcRun_ToDeleteContract, ARCtrl_ArcRun__ArcRun_ToCreateContract_6FCE9E49, ARCtrl_ArcRun__ArcRun_ToUpdateContract, ARCtrl_ArcRun__ArcRun_tryFromReadContract_Static_7570923F, ARCtrl_ArcRun__ArcRun_tryCWLFromReadContract_Static, ARCtrl_ArcRun__ArcRun_tryYMLFromReadContract_Static)
from ..arctrl_contract.arc_study import (ARCtrl_ArcStudy__ArcStudy_ToCreateContract_6FCE9E49, ARCtrl_ArcStudy__ArcStudy_ToUpdateContract, ARCtrl_ArcStudy__ArcStudy_tryFromReadContract_Static_7570923F)
from ..arctrl_contract.arc_workflow import (ARCtrl_ArcWorkflow__ArcWorkflow_ToDeleteContract, ARCtrl_ArcWorkflow__ArcWorkflow_ToCreateContract_6FCE9E49, ARCtrl_ArcWorkflow__ArcWorkflow_ToUpdateContract, ARCtrl_ArcWorkflow__ArcWorkflow_tryFromReadContract_Static_7570923F, ARCtrl_ArcWorkflow__ArcWorkflow_tryCWLFromReadContract_Static)
from ..arctrl_contract.contract import (Contract, DTOType, DTO)
from ..arctrl_contract.datamap import (ARCtrl_DataMap__DataMap_ToCreateContractForStudy_Z721C83C5, ARCtrl_DataMap__DataMap_ToUpdateContractForStudy_Z721C83C5, ARCtrl_DataMap__DataMap_ToCreateContractForAssay_Z721C83C5, ARCtrl_DataMap__DataMap_ToUpdateContractForAssay_Z721C83C5, ARCtrl_DataMap__DataMap_ToCreateContractForWorkflow_Z721C83C5, ARCtrl_DataMap__DataMap_ToUpdateContractForWorkflow_Z721C83C5, ARCtrl_DataMap__DataMap_ToCreateContractForRun_Z721C83C5, ARCtrl_DataMap__DataMap_ToUpdateContractForRun_Z721C83C5, ARCtrl_DataMap__DataMap_tryFromReadContractForAssay_Static, ARCtrl_DataMap__DataMap_tryFromReadContractForStudy_Static, ARCtrl_DataMap__DataMap_tryFromReadContractForWorkflow_Static, ARCtrl_DataMap__DataMap_tryFromReadContractForRun_Static)
from ..arctrl_contract.git import (Init_createInitContract_6DFDD678, gitignore_contract, gitattributes_contract, Init_createAddRemoteContract_Z721C83C5, Clone_createCloneContract_5000466F)
from ..arctrl_contract.validation_packages_config import (ValidationPackagesConfigHelper_ConfigFilePath, ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toCreateContract_Static_724DAE55, ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toDeleteContract_Static_724DAE55, ValidationPackagesConfigHelper_ReadContract, ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_tryFromReadContract_Static_7570923F)
from ..arctrl_core.arc_types import (ArcInvestigation, ArcAssay, ArcStudy, ArcRun, ArcWorkflow, ArcInvestigation_reflection)
from ..arctrl_core.comment import (Comment, Remark)
from ..arctrl_core.data import Data
from ..arctrl_core.data_context import DataContext
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.collections_ import (ResizeArray_iter, ResizeArray_map)
from ..arctrl_core.Helper.identifier import (create_missing_identifier, Study_fileNameFromIdentifier, Study_datamapFileNameFromIdentifier, Assay_fileNameFromIdentifier, Assay_datamapFileNameFromIdentifier, Workflow_fileNameFromIdentifier, Workflow_datamapFileNameFromIdentifier, Run_fileNameFromIdentifier, Run_datamapFileNameFromIdentifier)
from ..arctrl_core.identifier_setters import set_investigation_identifier
from ..arctrl_core.ontology_source_reference import OntologySourceReference
from ..arctrl_core.person import Person
from ..arctrl_core.publication import Publication
from ..arctrl_core.Table.arc_table import ArcTable
from ..arctrl_core.Table.arc_tables import ArcTables
from ..arctrl_core.Table.composite_cell import CompositeCell
from ..arctrl_core.Table.composite_column import CompositeColumn
from ..arctrl_cwl.cwlprocessing_unit import CWLProcessingUnit
from ..arctrl_cwl.parameter_reference import CWLParameterReference
from ..arctrl_file_system.file_system import FileSystem
from ..arctrl_file_system.file_system_tree import FileSystemTree
from ..arctrl_file_system.path import (get_assay_folder_path, get_run_folder_path, get_workflow_folder_path, get_study_folder_path)
from ..arctrl_json.encode import default_spaces
from ..arctrl_spreadsheet.arc_assay import ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F
from ..arctrl_spreadsheet.arc_investigation import ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF
from ..arctrl_spreadsheet.arc_run import ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8
from ..arctrl_spreadsheet.arc_study import ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522
from ..arctrl_spreadsheet.arc_workflow import ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E
from ..arctrl_spreadsheet.data_map import to_fs_workbook
from ..arctrl_validation_packages.validation_packages_config import ValidationPackagesConfig
from ..fable_library.array_ import (filter, map, choose, iterate as iterate_1, exists, fold, concat, contains as contains_2, append as append_1, try_pick, equals_with)
from ..fable_library.async_ import run_synchronously
from ..fable_library.async_builder import (Async, singleton)
from ..fable_library.list import FSharpList
from ..fable_library.map import of_seq as of_seq_1
from ..fable_library.map_util import add_to_dict
from ..fable_library.option import (value as value_9, default_arg, map as map_1, bind, default_arg_with)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.result import FSharpResult_2
from ..fable_library.seq import (to_array, contains, delay, append, singleton as singleton_1, iterate, try_find, find, empty, collect, to_list)
from ..fable_library.set import (of_seq, contains as contains_1, union_many, FSharpSet__Contains)
from ..fable_library.string_ import (starts_with_exact, join, to_fail, printf, replace, to_text)
from ..fable_library.types import (Array, FSharpRef)
from ..fable_library.util import (string_hash, IEnumerable_1, compare_primitives, curry2, ignore, safe_hash, get_enumerator, dispose, to_enumerable, equals)
from ..fs_spreadsheet.Cells.fs_cells_collection import Dictionary_tryGet
from ..thoth_json_core.types import IEncodable
from .ContractIO.contract_io import full_fill_contract_batch_async
from .ContractIO.file_system_helper import get_all_file_paths_async
from ..thoth_json_python.decode import Decode_fromString
from ..thoth_json_python.encode import to_string
from .license import License
from .rocrate_io import (ROCrate_get_decoderDeprecated, ROCrate_get_decoder, ROCrate_encoder_1E8A3F74)

def _expr4247() -> TypeInfo:
    return class_type("ARCtrl.ARC", None, ARC, ArcInvestigation_reflection())


class ARC(ArcInvestigation):
    def __init__(self, identifier: str, title: str | None=None, description: str | None=None, submission_date: str | None=None, public_release_date: str | None=None, ontology_source_references: Array[OntologySourceReference] | None=None, publications: Array[Publication] | None=None, contacts: Array[Person] | None=None, assays: Array[ArcAssay] | None=None, studies: Array[ArcStudy] | None=None, workflows: Array[ArcWorkflow] | None=None, runs: Array[ArcRun] | None=None, registered_study_identifiers: Array[str] | None=None, comments: Array[Comment] | None=None, remarks: Array[Remark] | None=None, fs: FileSystem | None=None, license: License | None=None) -> None:
        super().__init__(identifier, title, description, submission_date, public_release_date, ontology_source_references, publications, contacts, assays, studies, workflows, runs, registered_study_identifiers, comments, remarks)
        this: FSharpRef[ARC] = FSharpRef(None)
        this.contents = self
        self._license: License | None = license
        def _arrow4246(__unit: None=None) -> FileSystem:
            fs_1: FileSystem = default_arg(fs, FileSystem.create(tree = FileSystemTree(1, "", [])))
            return ARCAux_updateFSByARC(this.contents, license, fs_1)

        self._fs: FileSystem = _arrow4246()
        self.init_0040114: int = 1

    @property
    def FileSystem(self, __unit: None=None) -> FileSystem:
        this: ARC = self
        return this._fs

    @FileSystem.setter
    def FileSystem(self, fs: FileSystem) -> None:
        this: ARC = self
        this._fs = fs

    @property
    def License(self, __unit: None=None) -> License | None:
        this: ARC = self
        return this._license

    @License.setter
    def License(self, license: License | None=None) -> None:
        this: ARC = self
        this._license = license

    @staticmethod
    def from_arc_investigation(isa: ArcInvestigation, fs: FileSystem | None=None, license: License | None=None) -> ARC:
        return ARC(isa.Identifier, isa.Title, isa.Description, isa.SubmissionDate, isa.PublicReleaseDate, isa.OntologySourceReferences, isa.Publications, isa.Contacts, isa.Assays, isa.Studies, isa.Workflows, isa.Runs, isa.RegisteredStudyIdentifiers, isa.Comments, isa.Remarks, fs, license)

    def TryWriteAsync(self, arc_path: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetWriteContracts())

    def TryUpdateAsync(self, arc_path: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetUpdateContracts())

    def SetLicenseFulltext(self, fulltext: str, path: str | None=None) -> None:
        this: ARC = self
        match_value: License | None = this.License
        if match_value is None:
            license_1: License = License("fulltext", fulltext)
            this.License = license_1
            if path is not None:
                license_1.Path = value_9(path)


        else: 
            license: License = match_value
            license.Type = "fulltext"
            license.Content = fulltext
            if path is not None:
                license.Path = value_9(path)



    @staticmethod
    def try_load_async(arc_path: str) -> Async[FSharpResult_2[ARC, Array[str]]]:
        def _arrow4121(__unit: None=None) -> Async[FSharpResult_2[ARC, Array[str]]]:
            def _arrow4120(_arg: Array[str]) -> Async[FSharpResult_2[ARC, Array[str]]]:
                arc: ARC = ARC.from_file_paths(to_array(_arg))
                contracts: Array[Contract] = arc.GetReadContracts()
                def _arrow4119(_arg_1: FSharpResult_2[Array[Contract], Array[str]]) -> Async[FSharpResult_2[ARC, Array[str]]]:
                    ful_filled_contracts: FSharpResult_2[Array[Contract], Array[str]] = _arg_1
                    if ful_filled_contracts.tag == 1:
                        return singleton.Return(FSharpResult_2(1, ful_filled_contracts.fields[0]))

                    else: 
                        arc.SetISAFromContracts(ful_filled_contracts.fields[0])
                        return singleton.Return(FSharpResult_2(0, arc))


                return singleton.Bind(full_fill_contract_batch_async(arc_path, contracts), _arrow4119)

            return singleton.Bind(get_all_file_paths_async(arc_path), _arrow4120)

        return singleton.Delay(_arrow4121)

    def GetAssayRemoveContracts(self, assay_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4123:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4122(x: str, y: str) -> bool:
                    return x == y

                return _arrow4122

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(assay_identifier, this.AssayIdentifiers, ObjectExpr4123()):
            raise Exception("ARC does not contain assay with given name")

        assay: ArcAssay = this.GetAssay(assay_identifier)
        studies: Array[ArcStudy] = assay.StudiesRegisteredIn
        super().RemoveAssay(assay_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        assay_folder_path: str = get_assay_folder_path(assay_identifier)
        def predicate(p: str) -> bool:
            return not starts_with_exact(p, assay_folder_path)

        filtered_paths: Array[str] = filter(predicate, paths)
        this.SetFilePaths(filtered_paths)
        def _arrow4125(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4124(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(ARCtrl_ArcAssay__ArcAssay_ToDeleteContract(assay)), delay(_arrow4124))

        return to_array(delay(_arrow4125))

    def GetRunRemoveContracts(self, run_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4127:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4126(x: str, y: str) -> bool:
                    return x == y

                return _arrow4126

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(run_identifier, this.RunIdentifiers, ObjectExpr4127()):
            raise Exception("ARC does not contain run with given name")

        run: ArcRun = this.GetRun(run_identifier)
        this.DeleteRun(run_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        run_folder_path: str = get_run_folder_path(run_identifier)
        def predicate(p: str) -> bool:
            return not starts_with_exact(p, run_folder_path)

        filtered_paths: Array[str] = filter(predicate, paths)
        this.SetFilePaths(filtered_paths)
        def _arrow4130(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4129(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(ARCtrl_ArcRun__ArcRun_ToDeleteContract(run)), delay(_arrow4129))

        return to_array(delay(_arrow4130))

    def GetWorkflowRemoveContracts(self, workflow_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4132:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4131(x: str, y: str) -> bool:
                    return x == y

                return _arrow4131

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(workflow_identifier, this.WorkflowIdentifiers, ObjectExpr4132()):
            raise Exception("ARC does not contain workflow with given name")

        workflow: ArcWorkflow = this.GetWorkflow(workflow_identifier)
        this.DeleteWorkflow(workflow_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        workflow_folder_path: str = get_workflow_folder_path(workflow_identifier)
        def predicate(p: str) -> bool:
            return not starts_with_exact(p, workflow_folder_path)

        filtered_paths: Array[str] = filter(predicate, paths)
        this.SetFilePaths(filtered_paths)
        def _arrow4134(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4133(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(ARCtrl_ArcWorkflow__ArcWorkflow_ToDeleteContract(workflow)), delay(_arrow4133))

        return to_array(delay(_arrow4134))

    def TryRemoveAssayAsync(self, arc_path: str, assay_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetAssayRemoveContracts(assay_identifier))

    def TryRemoveRunAsync(self, arc_path: str, run_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetRunRemoveContracts(run_identifier))

    def TryRemoveWorkflowAsync(self, arc_path: str, workflow_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetWorkflowRemoveContracts(workflow_identifier))

    def RemoveRunAsync(self, arc_path: str, run_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4136(__unit: None=None) -> Async[None]:
            def _arrow4135(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not remove run, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRemoveRunAsync(arc_path, run_identifier), _arrow4135)

        return singleton.Delay(_arrow4136)

    def RemoveWorkflowAsync(self, arc_path: str, workflow_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4138(__unit: None=None) -> Async[None]:
            def _arrow4137(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not remove workflow, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRemoveWorkflowAsync(arc_path, workflow_identifier), _arrow4137)

        return singleton.Delay(_arrow4138)

    def TryRenameRunAsync(self, arc_path: str, old_run_identifier: str, new_run_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetRunRenameContracts(old_run_identifier, new_run_identifier))

    def TryRenameWorkflowAsync(self, arc_path: str, old_workflow_identifier: str, new_workflow_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetWorkflowRenameContracts(old_workflow_identifier, new_workflow_identifier))

    def RenameRunAsync(self, arc_path: str, old_run_identifier: str, new_run_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4140(__unit: None=None) -> Async[None]:
            def _arrow4139(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not rename run, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRenameRunAsync(arc_path, old_run_identifier, new_run_identifier), _arrow4139)

        return singleton.Delay(_arrow4140)

    def RenameWorkflowAsync(self, arc_path: str, old_workflow_identifier: str, new_workflow_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4142(__unit: None=None) -> Async[None]:
            def _arrow4141(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not rename workflow, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRenameWorkflowAsync(arc_path, old_workflow_identifier, new_workflow_identifier), _arrow4141)

        return singleton.Delay(_arrow4142)

    def GetAssayRenameContracts(self, old_assay_identifier: str, new_assay_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4144:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4143(x: str, y: str) -> bool:
                    return x == y

                return _arrow4143

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(old_assay_identifier, this.AssayIdentifiers, ObjectExpr4144()):
            raise Exception("ARC does not contain assay with given name")

        super().RenameAssay(old_assay_identifier, new_assay_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        old_assay_folder_path: str = get_assay_folder_path(old_assay_identifier)
        new_assay_folder_path: str = get_assay_folder_path(new_assay_identifier)
        def mapping(p: str) -> str:
            return replace(p, old_assay_folder_path, new_assay_folder_path)

        renamed_paths: Array[str] = map(mapping, paths, None)
        this.SetFilePaths(renamed_paths)
        def _arrow4146(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4145(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(Contract.create_rename(old_assay_folder_path, new_assay_folder_path)), delay(_arrow4145))

        return to_array(delay(_arrow4146))

    def GetRunRenameContracts(self, old_run_identifier: str, new_run_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4148:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4147(x: str, y: str) -> bool:
                    return x == y

                return _arrow4147

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(old_run_identifier, this.RunIdentifiers, ObjectExpr4148()):
            raise Exception("ARC does not contain run with given name")

        super().RenameRun(old_run_identifier, new_run_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        old_path: str = get_run_folder_path(old_run_identifier)
        new_path: str = get_run_folder_path(new_run_identifier)
        def mapping(p: str) -> str:
            return replace(p, old_path, new_path)

        renamed_paths: Array[str] = map(mapping, paths, None)
        this.SetFilePaths(renamed_paths)
        def _arrow4150(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4149(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(Contract.create_rename(old_path, new_path)), delay(_arrow4149))

        return to_array(delay(_arrow4150))

    def GetWorkflowRenameContracts(self, old_workflow_identifier: str, new_workflow_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4152:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4151(x: str, y: str) -> bool:
                    return x == y

                return _arrow4151

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(old_workflow_identifier, this.WorkflowIdentifiers, ObjectExpr4152()):
            raise Exception("ARC does not contain workflow with given name")

        super().RenameWorkflow(old_workflow_identifier, new_workflow_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        old_path: str = get_workflow_folder_path(old_workflow_identifier)
        new_path: str = get_workflow_folder_path(new_workflow_identifier)
        def mapping(p: str) -> str:
            return replace(p, old_path, new_path)

        renamed_paths: Array[str] = map(mapping, paths, None)
        this.SetFilePaths(renamed_paths)
        def _arrow4154(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4153(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(Contract.create_rename(old_path, new_path)), delay(_arrow4153))

        return to_array(delay(_arrow4154))

    def TryRenameAssayAsync(self, arc_path: str, old_assay_identifier: str, new_assay_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetAssayRenameContracts(old_assay_identifier, new_assay_identifier))

    def GetStudyRemoveContracts(self, study_identifier: str) -> Array[Contract]:
        this: ARC = self
        super().RemoveStudy(study_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        study_folder_path: str = get_study_folder_path(study_identifier)
        def predicate(p: str) -> bool:
            return not starts_with_exact(p, study_folder_path)

        filtered_paths: Array[str] = filter(predicate, paths)
        this.SetFilePaths(filtered_paths)
        return [Contract.create_delete(study_folder_path), ARCtrl_ArcInvestigation__ArcInvestigation_ToUpdateContract(this)]

    def TryRemoveStudyAsync(self, arc_path: str, study_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetStudyRemoveContracts(study_identifier))

    def GetStudyRenameContracts(self, old_study_identifier: str, new_study_identifier: str) -> Array[Contract]:
        this: ARC = self
        class ObjectExpr4156:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4155(x: str, y: str) -> bool:
                    return x == y

                return _arrow4155

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains(old_study_identifier, this.StudyIdentifiers, ObjectExpr4156()):
            raise Exception("ARC does not contain study with given name")

        super().RenameStudy(old_study_identifier, new_study_identifier)
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        old_study_folder_path: str = get_study_folder_path(old_study_identifier)
        new_study_folder_path: str = get_study_folder_path(new_study_identifier)
        def mapping(p: str) -> str:
            return replace(p, old_study_folder_path, new_study_folder_path)

        renamed_paths: Array[str] = map(mapping, paths, None)
        this.SetFilePaths(renamed_paths)
        def _arrow4158(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4157(__unit: None=None) -> IEnumerable_1[Contract]:
                return this.GetUpdateContracts()

            return append(singleton_1(Contract.create_rename(old_study_folder_path, new_study_folder_path)), delay(_arrow4157))

        return to_array(delay(_arrow4158))

    def TryRenameStudyAsync(self, arc_path: str, old_study_identifier: str, new_study_identifier: str) -> Async[FSharpResult_2[Array[Contract], Array[str]]]:
        this: ARC = self
        return full_fill_contract_batch_async(arc_path, this.GetStudyRenameContracts(old_study_identifier, new_study_identifier))

    def WriteAsync(self, arc_path: str) -> Async[None]:
        this: ARC = self
        def _arrow4160(__unit: None=None) -> Async[None]:
            def _arrow4159(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not write ARC, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryWriteAsync(arc_path), _arrow4159)

        return singleton.Delay(_arrow4160)

    def UpdateAsync(self, arc_path: str) -> Async[None]:
        this: ARC = self
        def _arrow4162(__unit: None=None) -> Async[None]:
            def _arrow4161(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not update ARC, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryUpdateAsync(arc_path), _arrow4161)

        return singleton.Delay(_arrow4162)

    def RemoveAssayAsync(self, arc_path: str, assay_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4164(__unit: None=None) -> Async[None]:
            def _arrow4163(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not remove assay, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRemoveAssayAsync(arc_path, assay_identifier), _arrow4163)

        return singleton.Delay(_arrow4164)

    def RenameAssayAsync(self, arc_path: str, old_assay_identifier: str, new_assay_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4166(__unit: None=None) -> Async[None]:
            def _arrow4165(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not rename assay, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRenameAssayAsync(arc_path, old_assay_identifier, new_assay_identifier), _arrow4165)

        return singleton.Delay(_arrow4166)

    def RemoveStudyAsync(self, arc_path: str, study_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4168(__unit: None=None) -> Async[None]:
            def _arrow4167(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not remove study, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRemoveStudyAsync(arc_path, study_identifier), _arrow4167)

        return singleton.Delay(_arrow4168)

    def RenameStudyAsync(self, arc_path: str, old_study_identifier: str, new_study_identifier: str) -> Async[None]:
        this: ARC = self
        def _arrow4170(__unit: None=None) -> Async[None]:
            def _arrow4169(_arg: FSharpResult_2[Array[Contract], Array[str]]) -> Async[None]:
                result: FSharpResult_2[Array[Contract], Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not rename study, failed with the following errors %s"))(appended)
                    return singleton.Zero()

                else: 
                    return singleton.Zero()


            return singleton.Bind(this.TryRenameStudyAsync(arc_path, old_study_identifier, new_study_identifier), _arrow4169)

        return singleton.Delay(_arrow4170)

    @staticmethod
    def load_async(arc_path: str) -> Async[ARC]:
        def _arrow4172(__unit: None=None) -> Async[ARC]:
            def _arrow4171(_arg: FSharpResult_2[ARC, Array[str]]) -> Async[ARC]:
                result: FSharpResult_2[ARC, Array[str]] = _arg
                if result.tag == 1:
                    def mapping(e: str) -> str:
                        return e

                    appended: str = join("\n", map(mapping, result.fields[0], None))
                    to_fail(printf("Could not load ARC, failed with the following errors %s"))(appended)
                    return singleton.Return(ARC(create_missing_identifier()))

                else: 
                    return singleton.Return(result.fields[0])


            return singleton.Bind(ARC.try_load_async(arc_path), _arrow4171)

        return singleton.Delay(_arrow4172)

    def Write(self, arc_path: str) -> None:
        this: ARC = self
        run_synchronously(this.WriteAsync(arc_path))

    def Update(self, arc_path: str) -> None:
        this: ARC = self
        run_synchronously(this.UpdateAsync(arc_path))

    def RemoveAssay(self, arc_path: str, assay_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RemoveAssayAsync(arc_path, assay_identifier))

    def RenameAssay(self, arc_path: str, old_assay_identifier: str, new_assay_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RenameAssayAsync(arc_path, old_assay_identifier, new_assay_identifier))

    def RemoveStudy(self, arc_path: str, study_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RemoveStudyAsync(arc_path, study_identifier))

    def RenameStudy(self, arc_path: str, old_study_identifier: str, new_study_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RenameStudyAsync(arc_path, old_study_identifier, new_study_identifier))

    def RemoveRun(self, arc_path: str, run_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RemoveRunAsync(arc_path, run_identifier))

    def RenameRun(self, arc_path: str, old_run_identifier: str, new_run_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RenameRunAsync(arc_path, old_run_identifier, new_run_identifier))

    def RemoveWorkflow(self, arc_path: str, workflow_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RemoveWorkflowAsync(arc_path, workflow_identifier))

    def RenameWorkflow(self, arc_path: str, old_workflow_identifier: str, new_workflow_identifier: str) -> None:
        this: ARC = self
        run_synchronously(this.RenameWorkflowAsync(arc_path, old_workflow_identifier, new_workflow_identifier))

    @staticmethod
    def load(arc_path: str) -> ARC:
        return run_synchronously(ARC.load_async(arc_path))

    def MakeDataFilesAbsolute(self, __unit: None=None) -> None:
        this: ARC = self
        class ObjectExpr4173:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        files_paths: Any = of_seq(this.FileSystem.Tree.ToFilePaths(), ObjectExpr4173())
        def check_existence_from_root(p: str) -> bool:
            return contains_1(p, files_paths)

        def update_column_option(data_name_function: Callable[[Data], str], col: CompositeColumn | None=None) -> None:
            (pattern_matching_result, col_2) = (None, None)
            if col is not None:
                def _arrow4174(__unit: None=None, data_name_function: Any=data_name_function, col: Any=col) -> bool:
                    col_1: CompositeColumn = col
                    return col_1.Header.IsDataColumn

                if _arrow4174():
                    pattern_matching_result = 0
                    col_2 = col

                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                def f(c: CompositeCell, data_name_function: Any=data_name_function, col: Any=col) -> None:
                    if c.AsData.FilePath is not None:
                        new_file_path: str = data_name_function(c.AsData)
                        c.AsData.FilePath = new_file_path


                ResizeArray_iter(f, col_2.Cells)

            elif pattern_matching_result == 1:
                pass


        def update_table(data_name_function_1: Callable[[Data], str], t: ArcTable) -> None:
            update_column_option(data_name_function_1, t.TryGetInputColumn())
            update_column_option(data_name_function_1, t.TryGetOutputColumn())

        def update_data_map(data_name_function_2: Callable[[Data], str], dm: DataMap) -> None:
            def action(c_1: DataContext, data_name_function_2: Any=data_name_function_2, dm: Any=dm) -> None:
                if c_1.FilePath is not None:
                    new_file_path_1: str = data_name_function_2(c_1)
                    c_1.FilePath = new_file_path_1


            iterate(action, dm.DataContexts)

        def action_2(s: ArcStudy) -> None:
            def f_1(d: Data, s: Any=s) -> str:
                return d.GetAbsolutePathForStudy(s.Identifier, check_existence_from_root)

            source_1: Array[ArcTable] = s.Tables
            iterate(curry2(update_table)(f_1), source_1)
            if s.DataMap is not None:
                update_data_map(f_1, value_9(s.DataMap))


        iterate(action_2, this.Studies)
        def action_4(a_1: ArcAssay) -> None:
            def f_2(d_1: Data, a_1: Any=a_1) -> str:
                return d_1.GetAbsolutePathForAssay(a_1.Identifier, check_existence_from_root)

            source_3: Array[ArcTable] = a_1.Tables
            iterate(curry2(update_table)(f_2), source_3)
            if a_1.DataMap is not None:
                update_data_map(f_2, value_9(a_1.DataMap))


        iterate(action_4, this.Assays)

    @staticmethod
    def from_file_paths(file_paths: Array[str]) -> ARC:
        fs: FileSystem = FileSystem.from_file_paths(file_paths)
        return ARC(create_missing_identifier(), None, None, None, None, None, None, None, None, None, None, None, None, None, None, fs)

    def SetFilePaths(self, file_paths: Array[str]) -> None:
        this: ARC = self
        tree: FileSystemTree = FileSystemTree.from_file_paths(file_paths)
        this._fs = FileSystem(tree, this._fs.History)

    def GetReadContracts(self, __unit: None=None) -> Array[Contract]:
        this: ARC = self
        return choose(try_isaread_contract_from_path, this._fs.Tree.ToFilePaths(), None)

    def SetISAFromContracts(self, contracts: Array[Contract]) -> None:
        this: ARC = self
        investigation: ArcInvestigation = ARCAux_getArcInvestigationFromContracts(contracts)
        ignore(set_investigation_identifier(investigation.Identifier, this))
        this.Title = investigation.Title
        this.Description = investigation.Description
        this.SubmissionDate = investigation.SubmissionDate
        this.PublicReleaseDate = investigation.PublicReleaseDate
        this.OntologySourceReferences = investigation.OntologySourceReferences
        this.Publications = investigation.Publications
        this.Contacts = investigation.Contacts
        this.Comments = investigation.Comments
        this.Remarks = investigation.Remarks
        this.RegisteredStudyIdentifiers = investigation.RegisteredStudyIdentifiers
        def mapping(tuple: tuple[ArcStudy, FSharpList[ArcAssay]]) -> ArcStudy:
            return tuple[0]

        studies: Array[ArcStudy] = map(mapping, ARCAux_getArcStudiesFromContracts(contracts), None)
        assays: Array[ArcAssay] = ARCAux_getArcAssaysFromContracts(contracts)
        workflows: Array[ArcWorkflow] = ARCAux_getArcWorkflowsFromContracts(contracts)
        runs: Array[ArcRun] = ARCAux_getArcRunsFromContracts(contracts)
        license: License | None = ARCAux_getLicenseFromContracts(contracts)
        def action(ai: str) -> None:
            def predicate(a: ArcAssay, ai: Any=ai) -> bool:
                return a.Identifier == ai

            if not exists(predicate, assays):
                this.DeleteAssay(ai)


        iterate_1(action, this.AssayIdentifiers)
        def action_1(si: str) -> None:
            def predicate_1(s: ArcStudy, si: Any=si) -> bool:
                return s.Identifier == si

            if not exists(predicate_1, studies):
                this.DeleteStudy(si)


        iterate_1(action_1, this.StudyIdentifiers)
        def action_2(study: ArcStudy) -> None:
            def predicate_2(s_1: ArcStudy, study: Any=study) -> bool:
                return s_1.Identifier == study.Identifier

            registered_study_opt: ArcStudy | None = try_find(predicate_2, this.Studies)
            if registered_study_opt is None:
                this.AddStudy(study)

            else: 
                registered_study: ArcStudy = registered_study_opt
                registered_study.UpdateReferenceByStudyFile(study, True)

            datamap: DataMap | None = ARCAux_getStudyDataMapFromContracts(study.Identifier, contracts)
            if study.DataMap is None:
                study.DataMap = datamap

            study.StaticHash = study.GetLightHashCode() or 0

        iterate_1(action_2, studies)
        def action_3(assay: ArcAssay) -> None:
            def predicate_3(a_1: ArcAssay, assay: Any=assay) -> bool:
                return a_1.Identifier == assay.Identifier

            registered_assay_opt: ArcAssay | None = try_find(predicate_3, this.Assays)
            if registered_assay_opt is None:
                this.AddAssay(assay)

            else: 
                registered_assay: ArcAssay = registered_assay_opt
                registered_assay.UpdateReferenceByAssayFile(assay, True)

            def predicate_4(a_2: ArcAssay, assay: Any=assay) -> bool:
                return a_2.Identifier == assay.Identifier

            assay_1: ArcAssay = find(predicate_4, this.Assays)
            updated_tables: ArcTables
            array_6: Array[ArcStudy] = assay_1.StudiesRegisteredIn
            def folder(tables: ArcTables, study_1: ArcStudy, assay: Any=assay) -> ArcTables:
                return ArcTables.update_reference_tables_by_sheets(ArcTables(study_1.Tables), tables, False)

            updated_tables = fold(folder, ArcTables(assay_1.Tables), array_6)
            datamap_1: DataMap | None = ARCAux_getAssayDataMapFromContracts(assay_1.Identifier, contracts)
            if assay_1.DataMap is None:
                assay_1.DataMap = datamap_1

            assay_1.Tables = updated_tables.Tables

        iterate_1(action_3, assays)
        def action_4(workflow: ArcWorkflow) -> None:
            datamap_2: DataMap | None = ARCAux_getWorkflowDataMapFromContracts(workflow.Identifier, contracts)
            cwl: CWLProcessingUnit | None = ARCAux_getWorkflowCWLFromContracts(workflow.Identifier, contracts)
            if workflow.DataMap is None:
                workflow.DataMap = datamap_2

            workflow.CWLDescription = cwl
            this.AddWorkflow(workflow)
            workflow.StaticHash = workflow.GetLightHashCode() or 0

        iterate_1(action_4, workflows)
        def action_5(run: ArcRun) -> None:
            datamap_3: DataMap | None = ARCAux_getRunDataMapFromContracts(run.Identifier, contracts)
            cwl_1: CWLProcessingUnit | None = ARCAux_getRunCWLFromContracts(run.Identifier, contracts)
            yml: Array[CWLParameterReference] | None = ARCAux_getRunYMLFromContracts(run.Identifier, contracts)
            if run.DataMap is None:
                run.DataMap = datamap_3

            run.CWLInput = default_arg(yml, [])
            run.CWLDescription = cwl_1
            this.AddRun(run)
            run.StaticHash = run.GetLightHashCode() or 0

        iterate_1(action_5, runs)
        def action_6(a_3: ArcAssay) -> None:
            a_3.StaticHash = a_3.GetLightHashCode() or 0

        iterate(action_6, this.Assays)
        def action_7(s_2: ArcStudy) -> None:
            s_2.StaticHash = s_2.GetLightHashCode() or 0

        iterate(action_7, this.Studies)
        this.License = license
        this.StaticHash = this.GetLightHashCode() or 0

    def UpdateFileSystem(self, __unit: None=None) -> None:
        this: ARC = self
        new_fs: FileSystem = ARCAux_updateFSByARC(this, this._license, this._fs)
        this._fs = new_fs

    def GetWriteContracts(self, skip_update_fs: bool | None=None) -> Array[Contract]:
        this: ARC = self
        if not default_arg(skip_update_fs, False):
            this.UpdateFileSystem()

        filemap: Any = dict([])
        add_to_dict(filemap, "isa.investigation.xlsx", (DTOType(4), DTO(0, ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(this))))
        this.StaticHash = this.GetLightHashCode() or 0
        def action(s: ArcStudy) -> None:
            s.StaticHash = s.GetLightHashCode() or 0
            add_to_dict(filemap, Study_fileNameFromIdentifier(s.Identifier), (DTOType(1), DTO(0, ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(s))))
            if s.DataMap is not None:
                dm: DataMap = value_9(s.DataMap)
                dm.StaticHash = safe_hash(dm) or 0
                add_to_dict(filemap, Study_datamapFileNameFromIdentifier(s.Identifier), (DTOType(5), DTO(0, to_fs_workbook(dm))))


        iterate(action, this.Studies)
        def action_1(a: ArcAssay) -> None:
            a.StaticHash = a.GetLightHashCode() or 0
            add_to_dict(filemap, Assay_fileNameFromIdentifier(a.Identifier), (DTOType(0), DTO(0, ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F(a))))
            if a.DataMap is not None:
                dm_1: DataMap = value_9(a.DataMap)
                dm_1.StaticHash = safe_hash(dm_1) or 0
                add_to_dict(filemap, Assay_datamapFileNameFromIdentifier(a.Identifier), (DTOType(5), DTO(0, to_fs_workbook(dm_1))))


        iterate(action_1, this.Assays)
        def action_2(w: ArcWorkflow) -> None:
            w.StaticHash = w.GetLightHashCode() or 0
            add_to_dict(filemap, Workflow_fileNameFromIdentifier(w.Identifier), (DTOType(2), DTO(0, ARCtrl_ArcWorkflow__ArcWorkflow_toFsWorkbook_Static_Z1C75CB0E(w))))
            if w.CWLDescription is not None:
                raise Exception("Not implemented yet: CWL description in ARC.GetWriteContracts")

            if w.DataMap is not None:
                dm_2: DataMap = value_9(w.DataMap)
                dm_2.StaticHash = safe_hash(dm_2) or 0
                add_to_dict(filemap, Workflow_datamapFileNameFromIdentifier(w.Identifier), (DTOType(5), DTO(0, to_fs_workbook(dm_2))))


        iterate(action_2, this.Workflows)
        def action_3(r: ArcRun) -> None:
            r.StaticHash = r.GetLightHashCode() or 0
            add_to_dict(filemap, Run_fileNameFromIdentifier(r.Identifier), (DTOType(3), DTO(0, ARCtrl_ArcRun__ArcRun_toFsWorkbook_Static_Z3EFAF6F8(r))))
            if r.CWLDescription is not None:
                raise Exception("Not implemented yet: CWL description in ARC.GetWriteContracts")

            if len(r.CWLInput) > 0:
                raise Exception("Not implemented yet: CWL YAML input in ARC.GetWriteContracts")

            if r.DataMap is not None:
                dm_3: DataMap = value_9(r.DataMap)
                dm_3.StaticHash = safe_hash(dm_3) or 0
                add_to_dict(filemap, Run_datamapFileNameFromIdentifier(r.Identifier), (DTOType(5), DTO(0, to_fs_workbook(dm_3))))


        iterate(action_3, this.Runs)
        match_value: License | None = this.License
        if match_value is None:
            pass

        else: 
            l: License = match_value
            match_value_1: str = l.Type
            l.StaticHash = safe_hash(l) or 0
            add_to_dict(filemap, l.Path, (DTOType(10), DTO(1, l.Content)))

        def mapping(fp: str) -> Contract:
            match_value_2: tuple[DTOType, DTO] | None = Dictionary_tryGet(fp, filemap)
            if match_value_2 is None:
                return Contract.create_create(fp, DTOType(10))

            else: 
                wb: DTO = match_value_2[1]
                dto: DTOType = match_value_2[0]
                return Contract.create_create(fp, dto, wb)


        return map(mapping, this._fs.Tree.ToFilePaths(True), None)

    def GetUpdateContracts(self, skip_update_fs: bool | None=None) -> Array[Contract]:
        this: ARC = self
        if this.StaticHash == 0:
            this.StaticHash = this.GetLightHashCode() or 0
            return this.GetWriteContracts(skip_update_fs)

        else: 
            def _arrow4205(__unit: None=None) -> IEnumerable_1[Contract]:
                hash_1: int = this.GetLightHashCode() or 0
                def _arrow4204(__unit: None=None) -> IEnumerable_1[Contract]:
                    this.StaticHash = hash_1 or 0
                    def _arrow4180(s: ArcStudy) -> IEnumerable_1[Contract]:
                        hash_2: int = s.GetLightHashCode() or 0
                        def _arrow4179(__unit: None=None) -> IEnumerable_1[Contract]:
                            s.StaticHash = hash_2 or 0
                            match_value: DataMap | None = s.DataMap
                            (pattern_matching_result, dm_2, dm_3) = (None, None, None)
                            if match_value is not None:
                                def _arrow4177(__unit: None=None) -> bool:
                                    dm: DataMap = match_value
                                    return dm.StaticHash == 0

                                if _arrow4177():
                                    pattern_matching_result = 0
                                    dm_2 = match_value

                                else: 
                                    def _arrow4178(__unit: None=None) -> bool:
                                        dm_1: DataMap = match_value
                                        return dm_1.StaticHash != safe_hash(dm_1)

                                    if _arrow4178():
                                        pattern_matching_result = 1
                                        dm_3 = match_value

                                    else: 
                                        pattern_matching_result = 2



                            else: 
                                pattern_matching_result = 2

                            if pattern_matching_result == 0:
                                def _arrow4175(__unit: None=None) -> IEnumerable_1[Contract]:
                                    dm_2.StaticHash = safe_hash(dm_2) or 0
                                    return empty()

                                return append(singleton_1(ARCtrl_DataMap__DataMap_ToCreateContractForStudy_Z721C83C5(dm_2, s.Identifier)), delay(_arrow4175))

                            elif pattern_matching_result == 1:
                                def _arrow4176(__unit: None=None) -> IEnumerable_1[Contract]:
                                    dm_3.StaticHash = safe_hash(dm_3) or 0
                                    return empty()

                                return append(singleton_1(ARCtrl_DataMap__DataMap_ToUpdateContractForStudy_Z721C83C5(dm_3, s.Identifier)), delay(_arrow4176))

                            elif pattern_matching_result == 2:
                                return empty()


                        return append(ARCtrl_ArcStudy__ArcStudy_ToCreateContract_6FCE9E49(s, True) if (s.StaticHash == 0) else (singleton_1(ARCtrl_ArcStudy__ArcStudy_ToUpdateContract(s)) if (s.StaticHash != hash_2) else empty()), delay(_arrow4179))

                    def _arrow4203(__unit: None=None) -> IEnumerable_1[Contract]:
                        def _arrow4186(a: ArcAssay) -> IEnumerable_1[Contract]:
                            hash_3: int = a.GetLightHashCode() or 0
                            def _arrow4185(__unit: None=None) -> IEnumerable_1[Contract]:
                                a.StaticHash = hash_3 or 0
                                match_value_1: DataMap | None = a.DataMap
                                (pattern_matching_result_1, dm_6, dm_7) = (None, None, None)
                                if match_value_1 is not None:
                                    def _arrow4183(__unit: None=None) -> bool:
                                        dm_4: DataMap = match_value_1
                                        return dm_4.StaticHash == 0

                                    if _arrow4183():
                                        pattern_matching_result_1 = 0
                                        dm_6 = match_value_1

                                    else: 
                                        def _arrow4184(__unit: None=None) -> bool:
                                            dm_5: DataMap = match_value_1
                                            return dm_5.StaticHash != safe_hash(dm_5)

                                        if _arrow4184():
                                            pattern_matching_result_1 = 1
                                            dm_7 = match_value_1

                                        else: 
                                            pattern_matching_result_1 = 2



                                else: 
                                    pattern_matching_result_1 = 2

                                if pattern_matching_result_1 == 0:
                                    def _arrow4181(__unit: None=None) -> IEnumerable_1[Contract]:
                                        dm_6.StaticHash = safe_hash(dm_6) or 0
                                        return empty()

                                    return append(singleton_1(ARCtrl_DataMap__DataMap_ToCreateContractForAssay_Z721C83C5(dm_6, a.Identifier)), delay(_arrow4181))

                                elif pattern_matching_result_1 == 1:
                                    def _arrow4182(__unit: None=None) -> IEnumerable_1[Contract]:
                                        dm_7.StaticHash = safe_hash(dm_7) or 0
                                        return empty()

                                    return append(singleton_1(ARCtrl_DataMap__DataMap_ToUpdateContractForAssay_Z721C83C5(dm_7, a.Identifier)), delay(_arrow4182))

                                elif pattern_matching_result_1 == 2:
                                    return empty()


                            return append(ARCtrl_ArcAssay__ArcAssay_ToCreateContract_6FCE9E49(a, True) if (a.StaticHash == 0) else (singleton_1(ARCtrl_ArcAssay__ArcAssay_ToUpdateContract(a)) if (a.StaticHash != hash_3) else empty()), delay(_arrow4185))

                        def _arrow4202(__unit: None=None) -> IEnumerable_1[Contract]:
                            def _arrow4192(w: ArcWorkflow) -> IEnumerable_1[Contract]:
                                hash_4: int = w.GetLightHashCode() or 0
                                def _arrow4191(__unit: None=None) -> IEnumerable_1[Contract]:
                                    w.StaticHash = hash_4 or 0
                                    match_value_2: DataMap | None = w.DataMap
                                    (pattern_matching_result_2, dm_10, dm_11) = (None, None, None)
                                    if match_value_2 is not None:
                                        def _arrow4189(__unit: None=None) -> bool:
                                            dm_8: DataMap = match_value_2
                                            return dm_8.StaticHash == 0

                                        if _arrow4189():
                                            pattern_matching_result_2 = 0
                                            dm_10 = match_value_2

                                        else: 
                                            def _arrow4190(__unit: None=None) -> bool:
                                                dm_9: DataMap = match_value_2
                                                return dm_9.StaticHash != safe_hash(dm_9)

                                            if _arrow4190():
                                                pattern_matching_result_2 = 1
                                                dm_11 = match_value_2

                                            else: 
                                                pattern_matching_result_2 = 2



                                    else: 
                                        pattern_matching_result_2 = 2

                                    if pattern_matching_result_2 == 0:
                                        def _arrow4187(__unit: None=None) -> IEnumerable_1[Contract]:
                                            dm_10.StaticHash = safe_hash(dm_10) or 0
                                            return empty()

                                        return append(singleton_1(ARCtrl_DataMap__DataMap_ToCreateContractForWorkflow_Z721C83C5(dm_10, w.Identifier)), delay(_arrow4187))

                                    elif pattern_matching_result_2 == 1:
                                        def _arrow4188(__unit: None=None) -> IEnumerable_1[Contract]:
                                            dm_11.StaticHash = safe_hash(dm_11) or 0
                                            return empty()

                                        return append(singleton_1(ARCtrl_DataMap__DataMap_ToUpdateContractForWorkflow_Z721C83C5(dm_11, w.Identifier)), delay(_arrow4188))

                                    elif pattern_matching_result_2 == 2:
                                        return empty()


                                return append(ARCtrl_ArcWorkflow__ArcWorkflow_ToCreateContract_6FCE9E49(w, True) if (w.StaticHash == 0) else (singleton_1(ARCtrl_ArcWorkflow__ArcWorkflow_ToUpdateContract(w)) if (w.StaticHash != hash_4) else empty()), delay(_arrow4191))

                            def _arrow4201(__unit: None=None) -> IEnumerable_1[Contract]:
                                def _arrow4198(r: ArcRun) -> IEnumerable_1[Contract]:
                                    hash_5: int = r.GetLightHashCode() or 0
                                    def _arrow4197(__unit: None=None) -> IEnumerable_1[Contract]:
                                        r.StaticHash = hash_5 or 0
                                        match_value_3: DataMap | None = r.DataMap
                                        (pattern_matching_result_3, dm_14, dm_15) = (None, None, None)
                                        if match_value_3 is not None:
                                            def _arrow4195(__unit: None=None) -> bool:
                                                dm_12: DataMap = match_value_3
                                                return dm_12.StaticHash == 0

                                            if _arrow4195():
                                                pattern_matching_result_3 = 0
                                                dm_14 = match_value_3

                                            else: 
                                                def _arrow4196(__unit: None=None) -> bool:
                                                    dm_13: DataMap = match_value_3
                                                    return dm_13.StaticHash != safe_hash(dm_13)

                                                if _arrow4196():
                                                    pattern_matching_result_3 = 1
                                                    dm_15 = match_value_3

                                                else: 
                                                    pattern_matching_result_3 = 2



                                        else: 
                                            pattern_matching_result_3 = 2

                                        if pattern_matching_result_3 == 0:
                                            def _arrow4193(__unit: None=None) -> IEnumerable_1[Contract]:
                                                dm_14.StaticHash = safe_hash(dm_14) or 0
                                                return empty()

                                            return append(singleton_1(ARCtrl_DataMap__DataMap_ToCreateContractForRun_Z721C83C5(dm_14, r.Identifier)), delay(_arrow4193))

                                        elif pattern_matching_result_3 == 1:
                                            def _arrow4194(__unit: None=None) -> IEnumerable_1[Contract]:
                                                dm_15.StaticHash = safe_hash(dm_15) or 0
                                                return empty()

                                            return append(singleton_1(ARCtrl_DataMap__DataMap_ToUpdateContractForRun_Z721C83C5(dm_15, r.Identifier)), delay(_arrow4194))

                                        elif pattern_matching_result_3 == 2:
                                            return empty()


                                    return append(ARCtrl_ArcRun__ArcRun_ToCreateContract_6FCE9E49(r, True) if (r.StaticHash == 0) else (singleton_1(ARCtrl_ArcRun__ArcRun_ToUpdateContract(r)) if (r.StaticHash != hash_5) else empty()), delay(_arrow4197))

                                def _arrow4200(__unit: None=None) -> IEnumerable_1[Contract]:
                                    match_value_4: License | None = this.License
                                    if match_value_4 is None:
                                        return empty()

                                    else: 
                                        l: License = match_value_4
                                        hash_6: int = safe_hash(l) or 0
                                        def _arrow4199(__unit: None=None) -> IEnumerable_1[Contract]:
                                            l.StaticHash = hash_6 or 0
                                            return empty()

                                        return append(singleton_1(l.ToCreateContract()) if (l.StaticHash == 0) else (singleton_1(l.ToUpdateContract()) if (l.StaticHash != hash_6) else empty()), delay(_arrow4199))


                                return append(collect(_arrow4198, this.Runs), delay(_arrow4200))

                            return append(collect(_arrow4192, this.Workflows), delay(_arrow4201))

                        return append(collect(_arrow4186, this.Assays), delay(_arrow4202))

                    return append(collect(_arrow4180, this.Studies), delay(_arrow4203))

                return append(singleton_1(ARCtrl_ArcInvestigation__ArcInvestigation_ToUpdateContract(this)) if (this.StaticHash != hash_1) else empty(), delay(_arrow4204))

            return to_array(delay(_arrow4205))


    def GetGitInitContracts(self, branch: str | None=None, repository_address: str | None=None, default_gitignore: bool | None=None, default_gitattributes: bool | None=None) -> Array[Contract]:
        default_gitignore_1: bool = default_arg(default_gitignore, False)
        default_gitattributes_1: bool = default_arg(default_gitattributes, False)
        def _arrow4209(__unit: None=None) -> IEnumerable_1[Contract]:
            def _arrow4208(__unit: None=None) -> IEnumerable_1[Contract]:
                def _arrow4207(__unit: None=None) -> IEnumerable_1[Contract]:
                    def _arrow4206(__unit: None=None) -> IEnumerable_1[Contract]:
                        return singleton_1(Init_createAddRemoteContract_Z721C83C5(value_9(repository_address))) if (repository_address is not None) else empty()

                    return append(singleton_1(gitattributes_contract) if default_gitattributes_1 else empty(), delay(_arrow4206))

                return append(singleton_1(gitignore_contract) if default_gitignore_1 else empty(), delay(_arrow4207))

            return append(singleton_1(Init_createInitContract_6DFDD678(branch)), delay(_arrow4208))

        return to_array(delay(_arrow4209))

    @staticmethod
    def get_clone_contract(remote_url: str, merge: bool | None=None, branch: str | None=None, token: tuple[str, str] | None=None, nolfs: bool | None=None) -> Contract:
        return Clone_createCloneContract_5000466F(remote_url, merge, branch, token, nolfs)

    def Copy(self, __unit: None=None) -> ARC:
        this: ARC = self
        next_assays: Array[ArcAssay] = []
        next_studies: Array[ArcStudy] = []
        next_workflows: Array[ArcWorkflow] = []
        next_runs: Array[ArcRun] = []
        enumerator: Any = get_enumerator(this.Assays)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                assay: ArcAssay = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                copy: ArcAssay = assay.Copy()
                (next_assays.append(copy))

        finally: 
            dispose(enumerator)

        enumerator_1: Any = get_enumerator(this.Studies)
        try: 
            while enumerator_1.System_Collections_IEnumerator_MoveNext():
                study: ArcStudy = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                copy_1: ArcStudy = study.Copy()
                (next_studies.append(copy_1))

        finally: 
            dispose(enumerator_1)

        enumerator_2: Any = get_enumerator(this.Workflows)
        try: 
            while enumerator_2.System_Collections_IEnumerator_MoveNext():
                workflow: ArcWorkflow = enumerator_2.System_Collections_Generic_IEnumerator_1_get_Current()
                copy_2: ArcWorkflow = workflow.Copy()
                (next_workflows.append(copy_2))

        finally: 
            dispose(enumerator_2)

        enumerator_3: Any = get_enumerator(this.Runs)
        try: 
            while enumerator_3.System_Collections_IEnumerator_MoveNext():
                run: ArcRun = enumerator_3.System_Collections_Generic_IEnumerator_1_get_Current()
                copy_3: ArcRun = run.Copy()
                (next_runs.append(copy_3))

        finally: 
            dispose(enumerator_3)

        def f(c: Comment) -> Comment:
            return c.Copy()

        next_comments: Array[Comment] = ResizeArray_map(f, this.Comments)
        def f_1(c_1: Remark) -> Remark:
            return c_1.Copy()

        next_remarks: Array[Remark] = ResizeArray_map(f_1, this.Remarks)
        def f_2(c_2: Person) -> Person:
            return c_2.Copy()

        next_contacts: Array[Person] = ResizeArray_map(f_2, this.Contacts)
        def f_3(c_3: Publication) -> Publication:
            return c_3.Copy()

        next_publications: Array[Publication] = ResizeArray_map(f_3, this.Publications)
        def f_4(c_4: OntologySourceReference) -> OntologySourceReference:
            return c_4.Copy()

        next_ontology_source_references: Array[OntologySourceReference] = ResizeArray_map(f_4, this.OntologySourceReferences)
        next_study_identifiers: Array[str] = list(this.RegisteredStudyIdentifiers)
        def mapping(_arg: License) -> License:
            return _arg.Copy()

        next_license: License | None = map_1(mapping, this.License)
        fs_copy: FileSystem = this._fs.Copy()
        return ARC(this.Identifier, this.Title, this.Description, this.SubmissionDate, this.PublicReleaseDate, next_ontology_source_references, next_publications, next_contacts, next_assays, next_studies, next_workflows, next_runs, next_study_identifiers, next_comments, next_remarks, fs_copy, next_license)

    def GetRegisteredPayload(self, IgnoreHidden: bool | None=None) -> FileSystemTree:
        this: ARC = self
        copy: ARC = this.Copy()
        registered_studies: Array[ArcStudy] = copy.Studies[:]
        def mapping(s: ArcStudy) -> Array[ArcAssay]:
            return s.RegisteredAssays[:]

        registered_assays: Array[ArcAssay] = concat(map(mapping, registered_studies, None), None)
        def _arrow4212(__unit: None=None) -> IEnumerable_1[str]:
            def _arrow4211(__unit: None=None) -> IEnumerable_1[str]:
                def _arrow4210(__unit: None=None) -> IEnumerable_1[str]:
                    return singleton_1(value_9(this.License).Path) if (this.License is not None) else empty()

                return append(singleton_1("README.md"), delay(_arrow4210))

            return append(singleton_1("isa.investigation.xlsx"), delay(_arrow4211))

        class ObjectExpr4213:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        def mapping_1(s_1: ArcStudy) -> Any:
            study_foldername: str = ((("" + "studies") + "/") + s_1.Identifier) + ""
            def _arrow4220(__unit: None=None, s_1: Any=s_1) -> IEnumerable_1[str]:
                def _arrow4219(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow4218(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow4217(table: ArcTable) -> IEnumerable_1[str]:
                            def _arrow4216(kv: Any) -> IEnumerable_1[str]:
                                text_value: str = kv[1].ToFreeTextCell().AsFreeText
                                def _arrow4215(__unit: None=None) -> IEnumerable_1[str]:
                                    def _arrow4214(__unit: None=None) -> IEnumerable_1[str]:
                                        return singleton_1(((((("" + study_foldername) + "/") + "protocols") + "/") + text_value) + "")

                                    return append(singleton_1(((((("" + study_foldername) + "/") + "resources") + "/") + text_value) + ""), delay(_arrow4214))

                                return append(singleton_1(text_value), delay(_arrow4215))

                            return collect(_arrow4216, table.Values)

                        return collect(_arrow4217, s_1.Tables)

                    return append(singleton_1(((("" + study_foldername) + "/") + "README.md") + ""), delay(_arrow4218))

                return append(singleton_1(((("" + study_foldername) + "/") + "isa.study.xlsx") + ""), delay(_arrow4219))

            class ObjectExpr4221:
                @property
                def Compare(self) -> Callable[[str, str], int]:
                    return compare_primitives

            return of_seq(to_list(delay(_arrow4220)), ObjectExpr4221())

        class ObjectExpr4222:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        def mapping_2(a: ArcAssay) -> Any:
            assay_foldername: str = ((("" + "assays") + "/") + a.Identifier) + ""
            def _arrow4229(__unit: None=None, a: Any=a) -> IEnumerable_1[str]:
                def _arrow4228(__unit: None=None) -> IEnumerable_1[str]:
                    def _arrow4227(__unit: None=None) -> IEnumerable_1[str]:
                        def _arrow4226(table_1: ArcTable) -> IEnumerable_1[str]:
                            def _arrow4225(kv_1: Any) -> IEnumerable_1[str]:
                                text_value_1: str = kv_1[1].ToFreeTextCell().AsFreeText
                                def _arrow4224(__unit: None=None) -> IEnumerable_1[str]:
                                    def _arrow4223(__unit: None=None) -> IEnumerable_1[str]:
                                        return singleton_1(((((("" + assay_foldername) + "/") + "protocols") + "/") + text_value_1) + "")

                                    return append(singleton_1(((((("" + assay_foldername) + "/") + "dataset") + "/") + text_value_1) + ""), delay(_arrow4223))

                                return append(singleton_1(text_value_1), delay(_arrow4224))

                            return collect(_arrow4225, table_1.Values)

                        return collect(_arrow4226, a.Tables)

                    return append(singleton_1(((("" + assay_foldername) + "/") + "README.md") + ""), delay(_arrow4227))

                return append(singleton_1(((("" + assay_foldername) + "/") + "isa.assay.xlsx") + ""), delay(_arrow4228))

            class ObjectExpr4230:
                @property
                def Compare(self) -> Callable[[str, str], int]:
                    return compare_primitives

            return of_seq(to_list(delay(_arrow4229)), ObjectExpr4230())

        class ObjectExpr4231:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        class ObjectExpr4232:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        include_files: Any = union_many(to_enumerable([of_seq(to_list(delay(_arrow4212)), ObjectExpr4213()), union_many(map(mapping_1, registered_studies, None), ObjectExpr4222()), union_many(map(mapping_2, registered_assays, None), ObjectExpr4231())]), ObjectExpr4232())
        ignore_hidden: bool = default_arg(IgnoreHidden, True)
        fs_copy: FileSystem = this._fs.Copy()
        def binder(tree_1: FileSystemTree) -> FileSystemTree | None:
            if ignore_hidden:
                def _arrow4233(n_1: str, tree_1: Any=tree_1) -> bool:
                    return not starts_with_exact(n_1, ".")

                return FileSystemTree.filter_folders(_arrow4233)(tree_1)

            else: 
                return tree_1


        def _arrow4235(__unit: None=None) -> FileSystemTree | None:
            tree: FileSystemTree
            def predicate(p: str) -> bool:
                if True if starts_with_exact(p, "workflows") else starts_with_exact(p, "runs"):
                    return True

                else: 
                    return FSharpSet__Contains(include_files, p)


            paths: Array[str] = filter(predicate, FileSystemTree.to_file_paths()(fs_copy.Tree))
            tree = FileSystemTree.from_file_paths(paths)
            def _arrow4234(n: str) -> bool:
                return not starts_with_exact(n, ".")

            return FileSystemTree.filter_files(_arrow4234)(tree) if ignore_hidden else tree

        return default_arg(bind(binder, _arrow4235()), FileSystemTree.from_file_paths([]))

    def GetAdditionalPayload(self, IgnoreHidden: bool | None=None) -> FileSystemTree:
        this: ARC = self
        ignore_hidden: bool = default_arg(IgnoreHidden, True)
        class ObjectExpr4236:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        registered_payload: Any = of_seq(FileSystemTree.to_file_paths()(this.GetRegisteredPayload()), ObjectExpr4236())
        def binder(tree_1: FileSystemTree) -> FileSystemTree | None:
            if ignore_hidden:
                def _arrow4237(n_1: str, tree_1: Any=tree_1) -> bool:
                    return not starts_with_exact(n_1, ".")

                return FileSystemTree.filter_folders(_arrow4237)(tree_1)

            else: 
                return tree_1


        def _arrow4239(__unit: None=None) -> FileSystemTree | None:
            tree: FileSystemTree
            def predicate(p: str) -> bool:
                return not FSharpSet__Contains(registered_payload, p)

            paths: Array[str] = filter(predicate, FileSystemTree.to_file_paths()(this._fs.Copy().Tree))
            tree = FileSystemTree.from_file_paths(paths)
            def _arrow4238(n: str) -> bool:
                return not starts_with_exact(n, ".")

            return FileSystemTree.filter_files(_arrow4238)(tree) if ignore_hidden else tree

        return default_arg(bind(binder, _arrow4239()), FileSystemTree.from_file_paths([]))

    @staticmethod
    def DefaultContracts() -> Any:
        class ObjectExpr4240:
            @property
            def Compare(self) -> Callable[[str, str], int]:
                return compare_primitives

        return of_seq_1(to_enumerable([(".gitignore", gitignore_contract), (".gitattributes", gitattributes_contract)]), ObjectExpr4240())

    @staticmethod
    def from_deprecated_rocrate_json_string(s: str) -> ARC:
        try: 
            s_1: str = replace(s, "bio:additionalProperty", "sdo:additionalProperty")
            isa: ArcInvestigation
            match_value: FSharpResult_2[ArcInvestigation, str] = Decode_fromString(ROCrate_get_decoderDeprecated(), s_1)
            if match_value.tag == 1:
                raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

            else: 
                isa = match_value.fields[0]

            return ARC.from_arc_investigation(isa)

        except Exception as ex:
            arg_1: str = str(ex)
            return to_fail(printf("Could not parse deprecated ARC-RO-Crate metadata: \n%s"))(arg_1)


    @staticmethod
    def from_rocrate_json_string(s: str) -> ARC:
        try: 
            pattern_input: tuple[ArcInvestigation, Array[str], License | None]
            match_value: FSharpResult_2[tuple[ArcInvestigation, Array[str], License | None], str] = Decode_fromString(ROCrate_get_decoder(), s)
            if match_value.tag == 1:
                raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

            else: 
                pattern_input = match_value.fields[0]

            file_system: FileSystem
            paths: Array[str] = list(pattern_input[1])
            file_system = FileSystem.from_file_paths(paths)
            return ARC.from_arc_investigation(pattern_input[0], file_system, pattern_input[2])

        except Exception as ex:
            arg_1: str = str(ex)
            return to_fail(printf("Could not parse ARC-RO-Crate metadata: \n%s"))(arg_1)


    def ToROCrateJsonString(self, spaces: int | None=None) -> str:
        this: ARC = self
        this.MakeDataFilesAbsolute()
        value: IEncodable = ROCrate_encoder_1E8A3F74(this, this._license, this._fs)
        return to_string(default_spaces(spaces), value)

    @staticmethod
    def to_rocrate_json_string(spaces: int | None=None) -> Callable[[ARC], str]:
        def _arrow4241(obj: ARC) -> str:
            return obj.ToROCrateJsonString(spaces)

        return _arrow4241

    def GetLicenseWriteContract(self, __unit: None=None) -> Contract:
        this: ARC = self
        def def_thunk(__unit: None=None) -> License:
            return License.GetDefaultLicense()

        _arg: License = default_arg_with(this.License, def_thunk)
        return _arg.ToCreateContract()

    def GetValidationPackagesConfigWriteContract(self, vpc: ValidationPackagesConfig) -> Contract:
        this: ARC = self
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        class ObjectExpr4243:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4242(x: str, y: str) -> bool:
                    return x == y

                return _arrow4242

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if not contains_2(ValidationPackagesConfigHelper_ConfigFilePath, paths, ObjectExpr4243()):
            file_paths: Array[str] = append_1([ValidationPackagesConfigHelper_ConfigFilePath], paths, None)
            this.SetFilePaths(file_paths)

        return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toCreateContract_Static_724DAE55(vpc)

    def GetValidationPackagesConfigDeleteContract(self, vpc: ValidationPackagesConfig) -> Contract:
        this: ARC = self
        paths: Array[str] = this.FileSystem.Tree.ToFilePaths()
        class ObjectExpr4245:
            @property
            def Equals(self) -> Callable[[str, str], bool]:
                def _arrow4244(x: str, y: str) -> bool:
                    return x == y

                return _arrow4244

            @property
            def GetHashCode(self) -> Callable[[str], int]:
                return string_hash

        if contains_2(ValidationPackagesConfigHelper_ConfigFilePath, paths, ObjectExpr4245()):
            def predicate(p: str) -> bool:
                return not (p == ValidationPackagesConfigHelper_ConfigFilePath)

            file_paths: Array[str] = filter(predicate, paths)
            this.SetFilePaths(file_paths)

        return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toDeleteContract_Static_724DAE55(vpc)

    def GetValidationPackagesConfigReadContract(self, __unit: None=None) -> Contract:
        return ValidationPackagesConfigHelper_ReadContract

    def GetValidationPackagesConfigFromReadContract(self, contract: Contract) -> ValidationPackagesConfig | None:
        return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_tryFromReadContract_Static_7570923F(contract)

    def ToFilePaths(self, remove_root: bool | None=None, skip_update_fs: bool | None=None) -> Array[str]:
        this: ARC = self
        if not default_arg(skip_update_fs, False):
            this.UpdateFileSystem()

        return this.FileSystem.Tree.ToFilePaths(remove_root)


ARC_reflection = _expr4247

def ARC__ctor_Z443FB9BF(identifier: str, title: str | None=None, description: str | None=None, submission_date: str | None=None, public_release_date: str | None=None, ontology_source_references: Array[OntologySourceReference] | None=None, publications: Array[Publication] | None=None, contacts: Array[Person] | None=None, assays: Array[ArcAssay] | None=None, studies: Array[ArcStudy] | None=None, workflows: Array[ArcWorkflow] | None=None, runs: Array[ArcRun] | None=None, registered_study_identifiers: Array[str] | None=None, comments: Array[Comment] | None=None, remarks: Array[Remark] | None=None, fs: FileSystem | None=None, license: License | None=None) -> ARC:
    return ARC(identifier, title, description, submission_date, public_release_date, ontology_source_references, publications, contacts, assays, studies, workflows, runs, registered_study_identifiers, comments, remarks, fs, license)


def ARCAux_getArcAssaysFromContracts(contracts: Array[Contract]) -> Array[ArcAssay]:
    def chooser(c: Contract, contracts: Any=contracts) -> ArcAssay | None:
        return ARCtrl_ArcAssay__ArcAssay_tryFromReadContract_Static_7570923F(c)

    return choose(chooser, contracts, None)


def ARCAux_getArcStudiesFromContracts(contracts: Array[Contract]) -> Array[tuple[ArcStudy, FSharpList[ArcAssay]]]:
    def chooser(c: Contract, contracts: Any=contracts) -> tuple[ArcStudy, FSharpList[ArcAssay]] | None:
        return ARCtrl_ArcStudy__ArcStudy_tryFromReadContract_Static_7570923F(c)

    return choose(chooser, contracts, None)


def ARCAux_getArcWorkflowsFromContracts(contracts: Array[Contract]) -> Array[ArcWorkflow]:
    def chooser(c: Contract, contracts: Any=contracts) -> ArcWorkflow | None:
        return ARCtrl_ArcWorkflow__ArcWorkflow_tryFromReadContract_Static_7570923F(c)

    return choose(chooser, contracts, None)


def ARCAux_getArcRunsFromContracts(contracts: Array[Contract]) -> Array[ArcRun]:
    def chooser(c: Contract, contracts: Any=contracts) -> ArcRun | None:
        return ARCtrl_ArcRun__ArcRun_tryFromReadContract_Static_7570923F(c)

    return choose(chooser, contracts, None)


def ARCAux_getAssayDataMapFromContracts(assay_identifier: str, contracts: Array[Contract]) -> DataMap | None:
    def chooser(c: Contract, assay_identifier: Any=assay_identifier, contracts: Any=contracts) -> DataMap | None:
        return ARCtrl_DataMap__DataMap_tryFromReadContractForAssay_Static(assay_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getStudyDataMapFromContracts(study_identifier: str, contracts: Array[Contract]) -> DataMap | None:
    def chooser(c: Contract, study_identifier: Any=study_identifier, contracts: Any=contracts) -> DataMap | None:
        return ARCtrl_DataMap__DataMap_tryFromReadContractForStudy_Static(study_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getWorkflowDataMapFromContracts(workflow_identifier: str, contracts: Array[Contract]) -> DataMap | None:
    def chooser(c: Contract, workflow_identifier: Any=workflow_identifier, contracts: Any=contracts) -> DataMap | None:
        return ARCtrl_DataMap__DataMap_tryFromReadContractForWorkflow_Static(workflow_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getRunDataMapFromContracts(run_identifier: str, contracts: Array[Contract]) -> DataMap | None:
    def chooser(c: Contract, run_identifier: Any=run_identifier, contracts: Any=contracts) -> DataMap | None:
        return ARCtrl_DataMap__DataMap_tryFromReadContractForRun_Static(run_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getWorkflowCWLFromContracts(workflow_identifier: str, contracts: Array[Contract]) -> CWLProcessingUnit | None:
    def chooser(c: Contract, workflow_identifier: Any=workflow_identifier, contracts: Any=contracts) -> CWLProcessingUnit | None:
        return ARCtrl_ArcWorkflow__ArcWorkflow_tryCWLFromReadContract_Static(workflow_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getRunCWLFromContracts(run_identifier: str, contracts: Array[Contract]) -> CWLProcessingUnit | None:
    def chooser(c: Contract, run_identifier: Any=run_identifier, contracts: Any=contracts) -> CWLProcessingUnit | None:
        return ARCtrl_ArcRun__ArcRun_tryCWLFromReadContract_Static(run_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getRunYMLFromContracts(run_identifier: str, contracts: Array[Contract]) -> Array[CWLParameterReference] | None:
    def chooser(c: Contract, run_identifier: Any=run_identifier, contracts: Any=contracts) -> Array[CWLParameterReference] | None:
        return ARCtrl_ArcRun__ArcRun_tryYMLFromReadContract_Static(run_identifier, c)

    return try_pick(chooser, contracts)


def ARCAux_getArcInvestigationFromContracts(contracts: Array[Contract]) -> ArcInvestigation:
    def chooser(c: Contract, contracts: Any=contracts) -> ArcInvestigation | None:
        return ARCtrl_ArcInvestigation__ArcInvestigation_tryFromReadContract_Static_7570923F(c)

    match_value: Array[ArcInvestigation] = choose(chooser, contracts, None)
    def _arrow4248(x: ArcInvestigation, y: ArcInvestigation, contracts: Any=contracts) -> bool:
        return equals(x, y)

    if (len(match_value) == 1) if (not equals_with(_arrow4248, match_value, None)) else False:
        return match_value[0]

    else: 
        arg: int = len(match_value) or 0
        return to_fail(printf("Could not find investigation in contracts. Expected exactly one investigation, but found %d."))(arg)



def ARCAux_getLicenseFromContracts(contracts: Array[Contract]) -> License | None:
    def chooser(c: Contract, contracts: Any=contracts) -> License | None:
        return License.try_from_read_contract(c)

    return try_pick(chooser, contracts)


def ARCAux_updateFSByARC(isa: ArcInvestigation, license: License | None, fs: FileSystem) -> FileSystem:
    assays_folder: FileSystemTree
    def mapping(a: ArcAssay, isa: Any=isa, license: Any=license, fs: Any=fs) -> FileSystemTree:
        return FileSystemTree.create_assay_folder(a.Identifier, a.DataMap is not None)

    assays: Array[FileSystemTree] = map(mapping, to_array(isa.Assays), None)
    assays_folder = FileSystemTree.create_assays_folder(assays)
    studies_folder: FileSystemTree
    def mapping_1(s: ArcStudy, isa: Any=isa, license: Any=license, fs: Any=fs) -> FileSystemTree:
        return FileSystemTree.create_study_folder(s.Identifier, s.DataMap is not None)

    studies: Array[FileSystemTree] = map(mapping_1, to_array(isa.Studies), None)
    studies_folder = FileSystemTree.create_studies_folder(studies)
    workflows_folder: FileSystemTree
    def mapping_2(w: ArcWorkflow, isa: Any=isa, license: Any=license, fs: Any=fs) -> FileSystemTree:
        return FileSystemTree.create_workflow_folder(w.Identifier, w.CWLDescription is not None, w.DataMap is not None)

    workflows: Array[FileSystemTree] = map(mapping_2, to_array(isa.Workflows), None)
    workflows_folder = FileSystemTree.create_workflows_folder(workflows)
    runs_folder: FileSystemTree
    def mapping_3(r: ArcRun, isa: Any=isa, license: Any=license, fs: Any=fs) -> FileSystemTree:
        return FileSystemTree.create_run_folder(r.Identifier, r.CWLDescription is not None, len(r.CWLInput) > 0, r.DataMap is not None)

    runs: Array[FileSystemTree] = map(mapping_3, to_array(isa.Runs), None)
    runs_folder = FileSystemTree.create_runs_folder(runs)
    investigation: FileSystemTree = FileSystemTree.create_investigation_file()
    tree_1: FileSystem
    def _arrow4254(__unit: None=None, isa: Any=isa, license: Any=license, fs: Any=fs) -> IEnumerable_1[FileSystemTree]:
        def _arrow4253(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
            def _arrow4252(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                def _arrow4251(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                    def _arrow4250(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                        def _arrow4249(__unit: None=None) -> IEnumerable_1[FileSystemTree]:
                            return singleton_1(FileSystemTree(0, value_9(license).Path)) if (license is not None) else empty()

                        return append(singleton_1(runs_folder), delay(_arrow4249))

                    return append(singleton_1(workflows_folder), delay(_arrow4250))

                return append(singleton_1(studies_folder), delay(_arrow4251))

            return append(singleton_1(assays_folder), delay(_arrow4252))

        return append(singleton_1(investigation), delay(_arrow4253))

    tree: FileSystemTree = FileSystemTree.create_root_folder(to_array(delay(_arrow4254)))
    tree_1 = FileSystem.create(tree = tree)
    return fs.Union(tree_1)


__all__ = ["ARC_reflection", "ARCAux_getArcAssaysFromContracts", "ARCAux_getArcStudiesFromContracts", "ARCAux_getArcWorkflowsFromContracts", "ARCAux_getArcRunsFromContracts", "ARCAux_getAssayDataMapFromContracts", "ARCAux_getStudyDataMapFromContracts", "ARCAux_getWorkflowDataMapFromContracts", "ARCAux_getRunDataMapFromContracts", "ARCAux_getWorkflowCWLFromContracts", "ARCAux_getRunCWLFromContracts", "ARCAux_getRunYMLFromContracts", "ARCAux_getArcInvestigationFromContracts", "ARCAux_getLicenseFromContracts", "ARCAux_updateFSByARC"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1597() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Assay.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Assay", string_type), ("ArcAssay", string_type), ("measurement_type", string_type), ("technology_type", string_type), ("technology_platform", string_type), ("data_files", string_type), ("materials", string_type), ("other_materials", string_type), ("samples", string_type), ("characteristic_categories", string_type), ("process_sequence", string_type), ("unit_categories", string_type), ("comments", string_type), ("filename", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Assay: str
    ArcAssay: str
    measurement_type: str
    technology_type: str
    technology_platform: str
    data_files: str
    materials: str
    other_materials: str
    samples: str
    characteristic_categories: str
    process_sequence: str
    unit_categories: str
    comments: str
    filename: str

IContext_reflection = _expr1597

def _arrow1612(__unit: None=None) -> IEncodable:
    class ObjectExpr1598(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1599(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:Dataset")

    class ObjectExpr1600(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:identifier")

    class ObjectExpr1601(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:additionalType")

    class ObjectExpr1602(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:variableMeasured")

    class ObjectExpr1603(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:measurementTechnique")

    class ObjectExpr1604(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:measurementMethod")

    class ObjectExpr1605(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:hasPart")

    class ObjectExpr1606(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            return helpers_8.encode_string("sdo:creator")

    class ObjectExpr1607(IEncodable):
        def Encode(self, helpers_9: IEncoderHelpers_1[Any]) -> Any:
            return helpers_9.encode_string("sdo:about")

    class ObjectExpr1608(IEncodable):
        def Encode(self, helpers_10: IEncoderHelpers_1[Any]) -> Any:
            return helpers_10.encode_string("sdo:comment")

    class ObjectExpr1610(IEncodable):
        def Encode(self, helpers_11: IEncoderHelpers_1[Any]) -> Any:
            return helpers_11.encode_string("sdo:url")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1598()), ("Assay", ObjectExpr1599()), ("identifier", ObjectExpr1600()), ("additionalType", ObjectExpr1601()), ("measurementType", ObjectExpr1602()), ("technologyType", ObjectExpr1603()), ("technologyPlatform", ObjectExpr1604()), ("dataFiles", ObjectExpr1605()), ("performers", ObjectExpr1606()), ("processSequence", ObjectExpr1607()), ("comments", ObjectExpr1608()), ("filename", ObjectExpr1610())])
    class ObjectExpr1611(IEncodable):
        def Encode(self, helpers_12: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_12))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_12.encode_object(arg)

    return ObjectExpr1611()


context_jsonvalue: IEncodable = _arrow1612()

__all__ = ["IContext_reflection", "context_jsonvalue"]





KeyPattern: str = ("^(?P<key>[^\\{\\[]+):\\s*(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

LineCommentPattern: str = ("^" + "\\<c f=(?P<comment>\\d+)\\/\\>") + "$"

ValuePattern: str = ("^(?P<value>.*?)\\s*?(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

InlineSequencePattern: str = ("^\\[(?P<inlineSequence>.+)\\]\\s*?(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

InlineJSONPattern: str = ("^\\{(?P<inlineSequence>.+)\\}\\s*?(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

SequenceOpenerPattern: str = ("^\\[\\s*(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

SequenceCloserPattern: str = ("^\\]\\s*(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

JSONOpenerPattern: str = ("^(?P<key>[^\\{\\[]+):\\s+\\{\\s*(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

JSONCloserPattern: str = ("^\\}\\s*(" + "\\<c f=(?P<comment>\\d+)\\/\\>") + ")?$"

__all__ = ["KeyPattern", "LineCommentPattern", "ValuePattern", "InlineSequencePattern", "InlineJSONPattern", "SequenceOpenerPattern", "SequenceCloserPattern", "JSONOpenerPattern", "JSONCloserPattern"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_validation_packages.validation_package import ValidationPackage
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.seq import (to_list, delay, collect, singleton)
from ..fable_library.util import (equals, IEnumerable_1)
from ..yamlicious.yamlicious_types import YAMLContent_create_27AED5E3
from ..yamlicious.decode import (object, IRequiredGetter, string as string_1, IOptionalGetter, IGetters)
from ..yamlicious.encode import (string, try_include)
from ..yamlicious.reader import read
from ..yamlicious.writer import write
from ..yamlicious.yamlicious_types import (YAMLElement, Config)
from .encode import default_whitespace

def ValidationPackage_encoder(validationpackage: ValidationPackage) -> YAMLElement:
    def chooser(tupled_arg: tuple[str, YAMLElement], validationpackage: Any=validationpackage) -> tuple[str, YAMLElement] | None:
        v: YAMLElement = tupled_arg[1]
        if equals(v, YAMLElement(5)):
            return None

        else: 
            return (tupled_arg[0], v)


    def _arrow3668(value: str, validationpackage: Any=validationpackage) -> YAMLElement:
        return string(value)

    obj_seq: FSharpList[tuple[str, YAMLElement]] = choose(chooser, of_array([("name", string(validationpackage.Name)), try_include("version", _arrow3668, validationpackage.Version)]))
    def _arrow3670(__unit: None=None, validationpackage: Any=validationpackage) -> IEnumerable_1[YAMLElement]:
        def _arrow3669(match_value: tuple[str, YAMLElement]) -> IEnumerable_1[YAMLElement]:
            return singleton(YAMLElement(0, YAMLContent_create_27AED5E3(match_value[0]), match_value[1]))

        return collect(_arrow3669, obj_seq)

    return YAMLElement(3, to_list(delay(_arrow3670)))


def _arrow3673(value_2: YAMLElement) -> ValidationPackage:
    def getter(get: IGetters) -> ValidationPackage:
        def _arrow3671(__unit: None=None, get: Any=get) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("name", string_1)

        def _arrow3672(__unit: None=None, get: Any=get) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("version", string_1)

        return ValidationPackage(_arrow3671(), _arrow3672())

    return object(getter, value_2)


ValidationPackage_decoder: Callable[[YAMLElement], ValidationPackage] = _arrow3673

def ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_fromYamlString_Static_Z721C83C5(s: str) -> ValidationPackage:
    return ValidationPackage_decoder(read(s))


def ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_toYamlString_Static_71136F3F(whitespace: int | None=None) -> Callable[[ValidationPackage], str]:
    def _arrow3675(vp: ValidationPackage, whitespace: Any=whitespace) -> str:
        element: YAMLElement = ValidationPackage_encoder(vp)
        whitespace_1: int = default_whitespace(whitespace) or 0
        def _arrow3674(c: Config) -> Config:
            return Config(whitespace_1, c.Level)

        return write(element, _arrow3674)

    return _arrow3675


def ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_toYamlString_71136F3F(this: ValidationPackage, whitespace: int | None=None) -> str:
    return ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_toYamlString_Static_71136F3F(whitespace)(this)


__all__ = ["ValidationPackage_encoder", "ValidationPackage_decoder", "ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_fromYamlString_Static_Z721C83C5", "ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_toYamlString_Static_71136F3F", "ARCtrl_ValidationPackages_ValidationPackage__ValidationPackage_toYamlString_71136F3F"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.arc_types import (ArcAssay, ArcStudy, ArcWorkflow, ArcRun, ArcInvestigation)
from ...arctrl_core.comment import Comment
from ...arctrl_core.Helper.collections_ import (Option_fromSeq, ResizeArray_map, ResizeArray_filter)
from ...arctrl_core.person import Person
from ...arctrl_core.publication import Publication
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.dataset import LDDataset
from ...fable_library.date import now
from ...fable_library.option import (bind, default_arg, map)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (to_list, delay, append)
from ...fable_library.types import Array
from ...fable_library.util import IEnumerable_1
from .assay import (AssayConversion_composeAssay_Z5C53FD5C, AssayConversion_decomposeAssay_Z6839B9E8)
from .basic import (BaseTypes_composeComment_Z13201A7E, BaseTypes_decomposeComment_Z2F770004)
from .date_time import (try_from_string, to_string)
from .person import (PersonConversion_composePerson_Z64D846DC, PersonConversion_decomposePerson_Z6839B9E8)
from .run import (RunConversion_composeRun_Z8CC08AC, RunConversion_decomposeRun_Z6839B9E8)
from .scholarly_article import (ScholarlyArticleConversion_composeScholarlyArticle_D324A6D, ScholarlyArticleConversion_decomposeScholarlyArticle_Z6839B9E8)
from .study import (StudyConversion_composeStudy_ZFE0E38E, StudyConversion_decomposeStudy_Z6839B9E8)
from .workflow import (WorkflowConversion_composeWorkflow_42450E6E, WorkflowConversion_decomposeWorkflow_Z6839B9E8)

def _expr3983() -> TypeInfo:
    return class_type("ARCtrl.Conversion.InvestigationConversion", None, InvestigationConversion)


class InvestigationConversion:
    ...

InvestigationConversion_reflection = _expr3983

def InvestigationConversion_composeInvestigation_5AEC717D(investigation: ArcInvestigation, fs: FileSystem | None=None) -> LDNode:
    name: str
    match_value: str | None = investigation.Title
    if match_value is None:
        raise Exception("Investigation must have a title")

    else: 
        name = match_value

    def _arrow3984(s: str, investigation: Any=investigation, fs: Any=fs) -> Any | None:
        return try_from_string(s)

    date_created: Any | None = bind(_arrow3984, investigation.SubmissionDate)
    def _arrow3985(s_1: str, investigation: Any=investigation, fs: Any=fs) -> Any | None:
        return try_from_string(s_1)

    date_published: Any = default_arg(bind(_arrow3985, investigation.PublicReleaseDate), now())
    def f(p: Publication, investigation: Any=investigation, fs: Any=fs) -> LDNode:
        return ScholarlyArticleConversion_composeScholarlyArticle_D324A6D(p)

    publications: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f, investigation.Publications))
    def f_1(c: Person, investigation: Any=investigation, fs: Any=fs) -> LDNode:
        return PersonConversion_composePerson_Z64D846DC(c)

    creators: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_1, investigation.Contacts))
    def f_2(c_1: Comment, investigation: Any=investigation, fs: Any=fs) -> LDNode:
        return BaseTypes_composeComment_Z13201A7E(c_1)

    comments: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_2, investigation.Comments))
    def _arrow3989(__unit: None=None, investigation: Any=investigation, fs: Any=fs) -> IEnumerable_1[LDNode]:
        def f_3(a_3: ArcAssay) -> LDNode:
            return AssayConversion_composeAssay_Z5C53FD5C(a_3, fs)

        def _arrow3988(__unit: None=None) -> IEnumerable_1[LDNode]:
            def f_4(s_2: ArcStudy) -> LDNode:
                return StudyConversion_composeStudy_ZFE0E38E(s_2, fs)

            def _arrow3987(__unit: None=None) -> IEnumerable_1[LDNode]:
                def f_5(w: ArcWorkflow) -> LDNode:
                    return WorkflowConversion_composeWorkflow_42450E6E(w, fs)

                def _arrow3986(__unit: None=None) -> IEnumerable_1[LDNode]:
                    def f_6(r: ArcRun) -> LDNode:
                        return RunConversion_composeRun_Z8CC08AC(r, fs)

                    return ResizeArray_map(f_6, investigation.Runs)

                return append(ResizeArray_map(f_5, investigation.Workflows), delay(_arrow3986))

            return append(ResizeArray_map(f_4, investigation.Studies), delay(_arrow3987))

        return append(ResizeArray_map(f_3, investigation.Assays), delay(_arrow3988))

    has_parts: Array[LDNode] | None = Option_fromSeq(list(to_list(delay(_arrow3989))))
    mentions: Array[LDNode] | None = Option_fromSeq([])
    return LDDataset.create_investigation(investigation.Identifier, name, None, creators, date_created, date_published, None, investigation.Description, has_parts, publications, comments, mentions)


def InvestigationConversion_decomposeInvestigation_Z6839B9E8(investigation: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcInvestigation:
    title: str | None
    match_value: str | None = LDDataset.try_get_name_as_string(investigation, context)
    title = LDDataset.try_get_headline_as_string(investigation, context) if (match_value is None) else match_value
    def _arrow3990(d: Any, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> str:
        return to_string(d)

    date_created: str | None = map(_arrow3990, LDDataset.try_get_date_created_as_date_time(investigation, context))
    def _arrow3991(d_1: Any, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> str:
        return to_string(d_1)

    date_published: str | None = map(_arrow3991, LDDataset.try_get_date_published_as_date_time(investigation, context))
    def f(p: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> Publication:
        return ScholarlyArticleConversion_decomposeScholarlyArticle_Z6839B9E8(p, graph, context)

    publications: Array[Publication] = ResizeArray_map(f, LDDataset.get_citations(investigation, graph, context))
    def f_1(c: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> Person:
        return PersonConversion_decomposePerson_Z6839B9E8(c, graph, context)

    creators: Array[Person] = ResizeArray_map(f_1, LDDataset.get_creators(investigation, graph, context))
    datasets: Array[LDNode] = LDDataset.get_has_parts_as_dataset(investigation, graph, context)
    def f_3(d_3: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> ArcStudy:
        return StudyConversion_decomposeStudy_Z6839B9E8(d_3, graph, context)

    def f_2(d_2: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> bool:
        return LDDataset.validate_study(d_2, context)

    studies: Array[ArcStudy] = ResizeArray_map(f_3, ResizeArray_filter(f_2, datasets))
    def f_5(d_5: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> ArcAssay:
        return AssayConversion_decomposeAssay_Z6839B9E8(d_5, graph, context)

    def f_4(d_4: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> bool:
        return LDDataset.validate_assay(d_4, context)

    assays: Array[ArcAssay] = ResizeArray_map(f_5, ResizeArray_filter(f_4, datasets))
    def f_7(d_7: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> ArcWorkflow:
        return WorkflowConversion_decomposeWorkflow_Z6839B9E8(d_7, graph, context)

    def f_6(d_6: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> bool:
        return LDDataset.validate_arcworkflow(d_6, graph, context)

    workflows: Array[ArcWorkflow] = ResizeArray_map(f_7, ResizeArray_filter(f_6, datasets))
    def f_9(d_9: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> ArcRun:
        return RunConversion_decomposeRun_Z6839B9E8(d_9, graph, context)

    def f_8(d_8: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> bool:
        return LDDataset.validate_arcrun(d_8, context)

    runs: Array[ArcRun] = ResizeArray_map(f_9, ResizeArray_filter(f_8, datasets))
    def f_10(c_1: LDNode, investigation: Any=investigation, graph: Any=graph, context: Any=context) -> Comment:
        return BaseTypes_decomposeComment_Z2F770004(c_1, context)

    comments: Array[Comment] = ResizeArray_map(f_10, LDDataset.get_comments(investigation, graph, context))
    return ArcInvestigation.create(LDDataset.get_identifier_as_string(investigation, context), title, LDDataset.try_get_description_as_string(investigation, context), date_created, date_published, None, publications, creators, assays, studies, workflows, runs, None, comments)


__all__ = ["InvestigationConversion_reflection", "InvestigationConversion_composeInvestigation_5AEC717D", "InvestigationConversion_decomposeInvestigation_Z6839B9E8"]




from __future__ import annotations
from abc import abstractmethod
from typing import Protocol

class IWidget(Protocol):
    @abstractmethod
    def close(self) -> None:
        ...


class IIntSlider(IWidget):
    @property
    @abstractmethod
    def max(self) -> int:
        ...

    @property
    @abstractmethod
    def min(self) -> int:
        ...

    @property
    @abstractmethod
    def value(self) -> int:
        ...


class IFloatSlider(IWidget):
    @property
    @abstractmethod
    def max(self) -> float:
        ...

    @property
    @abstractmethod
    def min(self) -> float:
        ...

    @property
    @abstractmethod
    def value(self) -> float:
        ...


class IExports(Protocol):
    pass




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.array_ import add_range_in_place
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import contains as contains_1
from ..fable_library.seq2 import Array_distinct
from ..fable_library.types import Array
from ..fable_library.util import (get_enumerator, dispose, equals, safe_hash, uncurry2)
from .Helper.collections_ import (ResizeArray_filter, ResizeArray_distinct, ResizeArray_collect, ResizeArray_append)
from .ontology_annotation import OntologyAnnotation
from .template import Template

def TemplatesAux_getComparer(match_all: bool | None=None) -> Callable[[bool, bool], bool]:
    if default_arg(match_all, False):
        def _arrow978(e: bool, match_all: Any=match_all) -> Callable[[bool], bool]:
            def _arrow977(e_1: bool) -> bool:
                return e and e_1

            return _arrow977

        return _arrow978

    else: 
        def _arrow980(e_2: bool, match_all: Any=match_all) -> Callable[[bool], bool]:
            def _arrow979(e_3: bool) -> bool:
                return e_2 or e_3

            return _arrow979

        return _arrow980



def TemplatesAux_filterOnTags(tag_getter: Callable[[Template], Array[OntologyAnnotation]], query_tags: Array[OntologyAnnotation], comparer: Callable[[bool, bool], bool], templates: Array[Template]) -> Array[Template]:
    def f(t: Template, tag_getter: Any=tag_getter, query_tags: Any=query_tags, comparer: Any=comparer, templates: Any=templates) -> bool:
        template_tags: Array[OntologyAnnotation] = tag_getter(t)
        is_valid: bool | None = None
        enumerator: Any = get_enumerator(query_tags)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                class ObjectExpr981:
                    @property
                    def Equals(self) -> Callable[[OntologyAnnotation, OntologyAnnotation], bool]:
                        return equals

                    @property
                    def GetHashCode(self) -> Callable[[OntologyAnnotation], int]:
                        return safe_hash

                contains: bool = contains_1(enumerator.System_Collections_Generic_IEnumerator_1_get_Current(), template_tags, ObjectExpr981())
                is_valid_1: bool | None = is_valid
                if is_valid_1 is not None:
                    maybe: bool = is_valid_1
                    is_valid = comparer(maybe, contains)

                else: 
                    is_valid = contains


        finally: 
            dispose(enumerator)

        return default_arg(is_valid, False)

    return ResizeArray_filter(f, templates)


def _expr989() -> TypeInfo:
    return class_type("ARCtrl.Templates", None, Templates)


class Templates:
    @staticmethod
    def get_distinct_tags(templates: Array[Template]) -> Array[OntologyAnnotation]:
        def f(t: Template) -> Array[OntologyAnnotation]:
            return t.Tags

        return ResizeArray_distinct(ResizeArray_collect(f, templates))

    @staticmethod
    def get_distinct_endpoint_repositories(templates: Array[Template]) -> Array[OntologyAnnotation]:
        def f(t: Template) -> Array[OntologyAnnotation]:
            return t.EndpointRepositories

        return ResizeArray_distinct(ResizeArray_collect(f, templates))

    @staticmethod
    def get_distinct_ontology_annotations(templates: Array[Template]) -> Array[OntologyAnnotation]:
        oas: Array[OntologyAnnotation] = []
        for idx in range(0, (len(templates) - 1) + 1, 1):
            t: Template = templates[idx]
            add_range_in_place(t.Tags, oas)
            add_range_in_place(t.EndpointRepositories, oas)
        class ObjectExpr982:
            @property
            def Equals(self) -> Callable[[OntologyAnnotation, OntologyAnnotation], bool]:
                return equals

            @property
            def GetHashCode(self) -> Callable[[OntologyAnnotation], int]:
                return safe_hash

        return Array_distinct(list(oas), ObjectExpr982())

    @staticmethod
    def filter_by_tags(query_tags: Array[OntologyAnnotation], match_all: bool | None=None) -> Callable[[Array[Template]], Array[Template]]:
        def _arrow984(templates: Array[Template]) -> Array[Template]:
            def _arrow983(t: Template) -> Array[OntologyAnnotation]:
                return t.Tags

            return TemplatesAux_filterOnTags(_arrow983, query_tags, uncurry2(TemplatesAux_getComparer(match_all)), templates)

        return _arrow984

    @staticmethod
    def filter_by_endpoint_repositories(query_tags: Array[OntologyAnnotation], match_all: bool | None=None) -> Callable[[Array[Template]], Array[Template]]:
        def _arrow986(templates: Array[Template]) -> Array[Template]:
            def _arrow985(t: Template) -> Array[OntologyAnnotation]:
                return t.EndpointRepositories

            return TemplatesAux_filterOnTags(_arrow985, query_tags, uncurry2(TemplatesAux_getComparer(match_all)), templates)

        return _arrow986

    @staticmethod
    def filter_by_ontology_annotation(query_tags: Array[OntologyAnnotation], match_all: bool | None=None) -> Callable[[Array[Template]], Array[Template]]:
        def _arrow988(templates: Array[Template]) -> Array[Template]:
            def _arrow987(t: Template) -> Array[OntologyAnnotation]:
                return ResizeArray_append(t.Tags, t.EndpointRepositories)

            return TemplatesAux_filterOnTags(_arrow987, query_tags, uncurry2(TemplatesAux_getComparer(match_all)), templates)

        return _arrow988

    @staticmethod
    def filter_by_data_plant(templates: Array[Template]) -> Array[Template]:
        def f(t: Template) -> bool:
            return t.Organisation.IsOfficial()

        return ResizeArray_filter(f, templates)


Templates_reflection = _expr989

__all__ = ["TemplatesAux_getComparer", "TemplatesAux_filterOnTags", "Templates_reflection"]




from __future__ import annotations

import os
import tempfile


def get_temp_file_name() -> str:
    fd, name = tempfile.mkstemp()
    os.close(fd)
    return name


def get_temp_path() -> str:
    return tempfile.gettempdir()


def get_extension(path: str) -> str:
    return os.path.splitext(path)[1]


def get_random_file_name() -> str:
    name = os.urandom(4).hex()
    ext = os.urandom(2).hex()[:3]
    return f"{name}.{ext}"


def get_file_name(path: str) -> str:
    return os.path.basename(path)


def get_directory_name(path: str) -> str:
    return os.path.dirname(path)


def get_file_name_without_extension(path: str) -> str:
    return os.path.splitext(os.path.basename(path))[0]


def get_full_path(path: str, base_path: str | None = None) -> str:
    return os.path.join(base_path or "", path)


def get_relative_path(relative_to: str, path: str) -> str:
    return os.path.relpath(path, relative_to)


def has_extension(path: str) -> bool:
    return os.path.splitext(path)[1] != ""


__all__ = [
    "get_temp_file_name",
    "get_temp_path",
    "get_extension",
    "get_random_file_name",
    "get_file_name",
    "get_directory_name",
    "get_file_name_without_extension",
    "get_full_path",
    "get_relative_path",
    "has_extension",
]



from __future__ import annotations
from typing import (Any, TypeVar)
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.rocrate_context import (init_v1_2, init_v1_1)
from ...fable_library.list import (FSharpList, is_empty, length, head)
from ...fable_library.result import FSharpResult_2
from ...fable_library.seq import (map, append, to_list, delay, singleton, empty)
from ...fable_library.types import Array
from ...fable_library.util import (IEnumerable_1, get_enumerator)
from ...thoth_json_core.decode import (Getters_2__ctor_Z4BE6C149, Getters_2, IRequiredGetter, string, Getters_2__get_Errors, resize_array)
from ...thoth_json_core.encode import seq
from ...thoth_json_core.types import (Decoder_1, ErrorReason_1, IDecoderHelpers_1, IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

class ObjectExpr1901(Decoder_1[LDContext]):
    def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any) -> FSharpResult_2[LDContext, tuple[str, ErrorReason_1[__A_]]]:
        this: Decoder_1[LDContext] = self
        if helpers.is_object(value):
            getters: Getters_2[__A_, Any] = Getters_2__ctor_Z4BE6C149(helpers, value)
            properties: IEnumerable_1[str] = helpers.get_properties(value)
            result: LDContext
            o: LDContext = LDContext()
            with get_enumerator(properties) as enumerator:
                while enumerator.System_Collections_IEnumerator_MoveNext():
                    property: str = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                    if (property != "@type") if (property != "@id") else False:
                        def _arrow1900(__unit: None=None) -> str:
                            object_arg: IRequiredGetter = getters.Required
                            return object_arg.Field(property, string)

                        o.AddMapping(property, _arrow1900())

            result = o
            match_value: FSharpList[tuple[str, ErrorReason_1[__A_]]] = Getters_2__get_Errors(getters)
            if not is_empty(match_value):
                errors: FSharpList[tuple[str, ErrorReason_1[__A_]]] = match_value
                return FSharpResult_2(1, ("", ErrorReason_1(7, errors))) if (length(errors) > 1) else FSharpResult_2(1, head(match_value))

            else: 
                return FSharpResult_2(0, result)


        elif helpers.is_string(value):
            s: str = helpers.as_string(value)
            return FSharpResult_2(0, init_v1_2()) if (True if (s == "https://w3id.org/ro/crate/1.2-DRAFT/context") else (s == "https://w3id.org/ro/crate/1.2/context")) else (FSharpResult_2(0, init_v1_1()) if (s == "https://w3id.org/ro/crate/1.1/context") else FSharpResult_2(1, ("", ErrorReason_1(0, "an object", value))))

        elif helpers.is_array(value):
            match_value_1: FSharpResult_2[Array[LDContext], tuple[str, ErrorReason_1[__A_]]] = resize_array(this).Decode(helpers, value)
            return FSharpResult_2(1, match_value_1.fields[0]) if (match_value_1.tag == 1) else FSharpResult_2(0, LDContext(None, match_value_1.fields[0]))

        else: 
            return FSharpResult_2(1, ("", ErrorReason_1(0, "an object", value)))



decoder: Decoder_1[LDContext] = ObjectExpr1901()

def encoder(ctx_mut: LDContext) -> IEncodable:
    while True:
        (ctx,) = (ctx_mut,)
        match_value: str | None = ctx.Name
        (pattern_matching_result,) = (None,)
        if match_value is not None:
            if match_value == "https://w3id.org/ro/crate/1.2-DRAFT/context":
                pattern_matching_result = 0

            elif match_value == "https://w3id.org/ro/crate/1.2/context":
                pattern_matching_result = 1

            elif match_value == "https://w3id.org/ro/crate/1.1/context":
                pattern_matching_result = 2

            else: 
                pattern_matching_result = 3


        else: 
            pattern_matching_result = 3

        if pattern_matching_result == 0:
            class ObjectExpr1902(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any], ctx: Any=ctx) -> Any:
                    return helpers.encode_string("https://w3id.org/ro/crate/1.2-DRAFT/context")

            return ObjectExpr1902()

        elif pattern_matching_result == 1:
            class ObjectExpr1903(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any], ctx: Any=ctx) -> Any:
                    return helpers_1.encode_string("https://w3id.org/ro/crate/1.2/context")

            return ObjectExpr1903()

        elif pattern_matching_result == 2:
            class ObjectExpr1904(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any], ctx: Any=ctx) -> Any:
                    return helpers_2.encode_string("https://w3id.org/ro/crate/1.1/context")

            return ObjectExpr1904()

        elif pattern_matching_result == 3:
            mappings: IEncodable
            def mapping(kv: Any, ctx: Any=ctx) -> tuple[str, IEncodable]:
                class ObjectExpr1905(IEncodable):
                    def Encode(self, helpers_3: IEncoderHelpers_1[Any], kv: Any=kv) -> Any:
                        return helpers_3.encode_string(kv[1])

                return (kv[0], ObjectExpr1905())

            values: IEnumerable_1[tuple[str, IEncodable]] = map(mapping, ctx.Mappings)
            class ObjectExpr1906(IEncodable):
                def Encode(self, helpers_4: IEncoderHelpers_1[Any], ctx: Any=ctx) -> Any:
                    def mapping_1(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                        return (tupled_arg[0], tupled_arg[1].Encode(helpers_4))

                    arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_1, values)
                    return helpers_4.encode_object(arg)

            mappings = ObjectExpr1906()
            if len(ctx.BaseContexts) == 0:
                return mappings

            elif (len(ctx.Mappings) == 0) if (len(ctx.BaseContexts) == 1) else False:
                ctx_mut = ctx.BaseContexts[0]
                continue

            else: 
                def _arrow1908(__unit: None=None, ctx: Any=ctx) -> IEnumerable_1[IEncodable]:
                    source_4: IEnumerable_1[IEncodable] = map(encoder, ctx.BaseContexts)
                    def _arrow1907(__unit: None=None) -> IEnumerable_1[IEncodable]:
                        return singleton(mappings) if (len(ctx.Mappings) != 0) else empty()

                    return append(to_list(delay(_arrow1907)), source_4)

                return seq(_arrow1908())


        break


__all__ = ["decoder", "encoder"]




from __future__ import annotations
from typing import Any
from ..arctrl_core.arc_types import (ArcStudy, ArcAssay)
from ..arctrl_core.Helper.identifier import Study_fileNameFromIdentifier
from ..arctrl_file_system.file_system_tree import FileSystemTree
from ..arctrl_file_system.path import (combine_many, get_study_folder_path)
from ..arctrl_spreadsheet.arc_study import (ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522, ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D)
from ..fable_library.array_ import equals_with
from ..fable_library.list import FSharpList
from ..fable_library.option import default_arg
from ..fable_library.seq import (to_array, delay, append, collect, singleton, empty)
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from .contract import (Contract, DTOType, DTO)

def _007CStudyPath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3728(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 3) if (not equals_with(_arrow3728, input, None)) else False:
        if input[0] == "studies":
            if input[2] == "isa.study.xlsx":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        any_study_name: str = input[1]
        return combine_many(input)

    elif pattern_matching_result == 1:
        return None



def ARCtrl_ArcStudy__ArcStudy_ToCreateContract_6FCE9E49(this: ArcStudy, WithFolder: bool | None=None) -> Array[Contract]:
    with_folder: bool = default_arg(WithFolder, False)
    path: str = Study_fileNameFromIdentifier(this.Identifier)
    c: Contract = Contract.create_create(path, DTOType(1), DTO(0, ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(this)))
    def _arrow3732(__unit: None=None, this: Any=this, WithFolder: Any=WithFolder) -> IEnumerable_1[Contract]:
        def _arrow3730(__unit: None=None) -> IEnumerable_1[Contract]:
            folder_fs: FileSystemTree = FileSystemTree.create_studies_folder([FileSystemTree.create_study_folder(this.Identifier)])
            def _arrow3729(p: str) -> IEnumerable_1[Contract]:
                return singleton(Contract.create_create(p, DTOType(10))) if ((p != "studies/.gitkeep") if (p != path) else False) else empty()

            return collect(_arrow3729, folder_fs.ToFilePaths(False))

        def _arrow3731(__unit: None=None) -> IEnumerable_1[Contract]:
            return singleton(c)

        return append(_arrow3730() if with_folder else empty(), delay(_arrow3731))

    return to_array(delay(_arrow3732))


def ARCtrl_ArcStudy__ArcStudy_ToUpdateContract(this: ArcStudy) -> Contract:
    path: str = Study_fileNameFromIdentifier(this.Identifier)
    return Contract.create_update(path, DTOType(1), DTO(0, ARCtrl_ArcStudy__ArcStudy_toFsWorkbook_Static_Z4CEFA522(this)))


def ARCtrl_ArcStudy__ArcStudy_ToDeleteContract(this: ArcStudy) -> Contract:
    path: str = get_study_folder_path(this.Identifier)
    return Contract.create_delete(path)


def ARCtrl_ArcStudy__ArcStudy_toDeleteContract_Static_1680536E(study: ArcStudy) -> Contract:
    return ARCtrl_ArcStudy__ArcStudy_ToDeleteContract(study)


def ARCtrl_ArcStudy__ArcStudy_toCreateContract_Static_Z76BBA099(study: ArcStudy, WithFolder: bool | None=None) -> Array[Contract]:
    return ARCtrl_ArcStudy__ArcStudy_ToCreateContract_6FCE9E49(study, WithFolder)


def ARCtrl_ArcStudy__ArcStudy_toUpdateContract_Static_1680536E(study: ArcStudy) -> Contract:
    return ARCtrl_ArcStudy__ArcStudy_ToUpdateContract(study)


def ARCtrl_ArcStudy__ArcStudy_tryFromReadContract_Static_7570923F(c: Contract) -> tuple[ArcStudy, FSharpList[ArcAssay]] | None:
    (pattern_matching_result, fsworkbook) = (None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 1:
                if c.DTO is not None:
                    if c.DTO.tag == 0:
                        pattern_matching_result = 0
                        fsworkbook = c.DTO.fields[0]

                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return ARCtrl_ArcStudy__ArcStudy_fromFsWorkbook_Static_32154C9D(fsworkbook)

    elif pattern_matching_result == 1:
        return None



__all__ = ["_007CStudyPath_007C__007C", "ARCtrl_ArcStudy__ArcStudy_ToCreateContract_6FCE9E49", "ARCtrl_ArcStudy__ArcStudy_ToUpdateContract", "ARCtrl_ArcStudy__ArcStudy_ToDeleteContract", "ARCtrl_ArcStudy__ArcStudy_toDeleteContract_Static_1680536E", "ARCtrl_ArcStudy__ArcStudy_toCreateContract_Static_Z76BBA099", "ARCtrl_ArcStudy__ArcStudy_toUpdateContract_Static_1680536E", "ARCtrl_ArcStudy__ArcStudy_tryFromReadContract_Static_7570923F"]




from __future__ import annotations
from dataclasses import dataclass
from typing import Any
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..dynamic_obj.dyn_obj import set_optional_property
from ..fable_library.reflection import (TypeInfo, string_type, option_type, class_type, record_type, array_type, float64_type, union_type)
from ..fable_library.types import (Record, FSharpRef, Array, Union)
from .cwltypes import (SchemaDefRequirementType_reflection, SoftwarePackage_reflection, CWLType_reflection)

def _expr482() -> TypeInfo:
    return record_type("ARCtrl.CWL.DockerRequirement", [], DockerRequirement, lambda: [("DockerPull", option_type(string_type)), ("DockerFile", option_type(class_type("Microsoft.FSharp.Collections.FSharpMap`2", [string_type, string_type]))), ("DockerImageId", option_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class DockerRequirement(Record):
    DockerPull: str | None
    DockerFile: Any | None
    DockerImageId: str | None

DockerRequirement_reflection = _expr482

def DockerRequirement_create_Z3FC43B23(docker_pull: str | None=None, docker_file: Any | None=None, docker_image_id: str | None=None) -> DockerRequirement:
    return DockerRequirement(docker_pull, docker_file, docker_image_id)


def _expr483() -> TypeInfo:
    return record_type("ARCtrl.CWL.EnvironmentDef", [], EnvironmentDef, lambda: [("EnvName", string_type), ("EnvValue", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class EnvironmentDef(Record):
    EnvName: str
    EnvValue: str

EnvironmentDef_reflection = _expr483

def _expr484() -> TypeInfo:
    return class_type("ARCtrl.CWL.ResourceRequirementInstance", None, ResourceRequirementInstance, DynamicObj_reflection())


class ResourceRequirementInstance(DynamicObj):
    def __init__(self, cores_min: Any | None=None, cores_max: Any | None=None, ram_min: Any | None=None, ram_max: Any | None=None, tmpdir_min: Any | None=None, tmpdir_max: Any | None=None, outdir_min: Any | None=None, outdir_max: Any | None=None) -> None:
        super().__init__()
        this: FSharpRef[ResourceRequirementInstance] = FSharpRef(None)
        this.contents = self
        self.init_004031: int = 1
        set_optional_property("coresMin", cores_min, this.contents)
        set_optional_property("coresMax", cores_max, this.contents)
        set_optional_property("ramMin", ram_min, this.contents)
        set_optional_property("ramMax", ram_max, this.contents)
        set_optional_property("tmpdirMin", tmpdir_min, this.contents)
        set_optional_property("tmpdirMax", tmpdir_max, this.contents)
        set_optional_property("outdirMin", outdir_min, this.contents)
        set_optional_property("outdirMax", outdir_max, this.contents)


ResourceRequirementInstance_reflection = _expr484

def ResourceRequirementInstance__ctor_D76FC00(cores_min: Any | None=None, cores_max: Any | None=None, ram_min: Any | None=None, ram_max: Any | None=None, tmpdir_min: Any | None=None, tmpdir_max: Any | None=None, outdir_min: Any | None=None, outdir_max: Any | None=None) -> ResourceRequirementInstance:
    return ResourceRequirementInstance(cores_min, cores_max, ram_min, ram_max, tmpdir_min, tmpdir_max, outdir_min, outdir_max)


def _expr485() -> TypeInfo:
    return union_type("ARCtrl.CWL.Requirement", [], Requirement, lambda: [[], [("Item", array_type(SchemaDefRequirementType_reflection()))], [("Item", DockerRequirement_reflection())], [("Item", array_type(SoftwarePackage_reflection()))], [("Item", array_type(CWLType_reflection()))], [("Item", array_type(EnvironmentDef_reflection()))], [], [("Item", ResourceRequirementInstance_reflection())], [], [], [], [("Item", float64_type)], [], [], [], []])


class Requirement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["InlineJavascriptRequirement", "SchemaDefRequirement", "DockerRequirement", "SoftwareRequirement", "InitialWorkDirRequirement", "EnvVarRequirement", "ShellCommandRequirement", "ResourceRequirement", "WorkReuseRequirement", "NetworkAccessRequirement", "InplaceUpdateRequirement", "ToolTimeLimitRequirement", "SubworkflowFeatureRequirement", "ScatterFeatureRequirement", "MultipleInputFeatureRequirement", "StepInputExpressionRequirement"]


Requirement_reflection = _expr485

__all__ = ["DockerRequirement_reflection", "DockerRequirement_create_Z3FC43B23", "EnvironmentDef_reflection", "ResourceRequirementInstance_reflection", "Requirement_reflection"]

